# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

DEFAULT_RAY_PORT = 6379
DEFAULT_RAY_DASHBOARD_PORT = 8265
DEFAULT_RAY_TEMP_DIR = "/tmp/ray"  # noqa: S108
DEFAULT_RAY_METRICS_PORT = 8080
DEFAULT_RAY_DASHBOARD_HOST = "127.0.0.1"
DEFAULT_RAY_CLIENT_SERVER_PORT = 10001
DEFAULT_RAY_AUTOSCALER_METRIC_PORT = 44217
DEFAULT_RAY_DASHBOARD_METRIC_PORT = 44227

# We cannot use a free port between 10000 and 19999 as it is used by Ray.
DEFAULT_RAY_MIN_WORKER_PORT = 10002
DEFAULT_RAY_MAX_WORKER_PORT = 19999



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass

os.environ["RAPIDS_NO_INITIALIZE"] = "1"

import numpy as np
import pandas as pd
import torch
from transformers import AutoModelForSequenceClassification

from nemo_curator.stages.base import CompositeStage, ProcessingStage
from nemo_curator.stages.text.models.model import ModelStage
from nemo_curator.stages.text.models.tokenizer import TokenizerStage
from nemo_curator.stages.text.models.utils import ATTENTION_MASK_COLUMN, INPUT_ID_COLUMN, format_name_with_suffix
from nemo_curator.stages.text.modules.score_filter import Filter
from nemo_curator.tasks import DocumentBatch

from .constants import DEBERTA_TOKENIZER_PADDING_SIDE

FINEWEB_EDU_MODEL_IDENTIFIER = "HuggingFaceFW/fineweb-edu-classifier"
FINEWEB_MIXTRAL_EDU_MODEL_IDENTIFIER = "nvidia/nemocurator-fineweb-mixtral-edu-classifier"
FINEWEB_NEMOTRON_EDU_MODEL_IDENTIFIER = "nvidia/nemocurator-fineweb-nemotron-4-edu-classifier"
MAX_SEQ_LENGTH = 512


class FineWebModelStage(ModelStage):
    """
    Stage for Hugging Face model inference.

    Args:
        model_identifier: The identifier of the Hugging Face model.
        cache_dir: The Hugging Face cache directory. Defaults to None.
        pred_column: The name of the prediction column.
        float_score_column: The name of the float score column.
        int_score_column: The name of the integer score column.
        model_inference_batch_size: The size of the batch for model inference. Defaults to 256.
        has_seq_order: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        autocast: Whether to use autocast. When True, we trade off minor accuracy for faster inference.
            Defaults to True.

    """

    def __init__(  # noqa: PLR0913
        self,
        model_identifier: str,
        cache_dir: str | None = None,
        pred_column: str = "preds",
        float_score_column: str = "float_score",
        int_score_column: str = "int_score",
        model_inference_batch_size: int = 256,
        has_seq_order: bool = True,
        autocast: bool = True,
    ):
        super().__init__(
            model_identifier=model_identifier,
            cache_dir=cache_dir,
            has_seq_order=has_seq_order,
            model_inference_batch_size=model_inference_batch_size,
            padding_side=DEBERTA_TOKENIZER_PADDING_SIDE,
            unpack_inference_batch=True,
        )

        self.pred_column = pred_column
        self.float_score_column = float_score_column
        self.int_score_column = int_score_column
        self.autocast = autocast

    def outputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [self.pred_column, self.float_score_column, self.int_score_column]

    @staticmethod
    def configure_forward(model: torch.nn.Module) -> torch.nn.Module:
        original_forward = model.forward

        @torch.no_grad()
        def custom_forward(*args, **kwargs) -> torch.Tensor:
            output = original_forward(*args, **kwargs)
            del args, kwargs
            return output.logits.squeeze(-1).float()

        model.forward = custom_forward
        return model

    def _setup(self, local_files_only: bool = True) -> None:
        model = AutoModelForSequenceClassification.from_pretrained(
            self.model_identifier,
            cache_dir=self.cache_dir,
            local_files_only=local_files_only,
        ).cuda()
        self.model = self.configure_forward(model)

    def process_model_output(
        self, outputs: torch.Tensor, _: dict[str, torch.Tensor] | None = None
    ) -> dict[str, np.ndarray]:
        logits = outputs.cpu().numpy()

        float_scores = logits.tolist()
        float_scores = [min(5.0, max(0.0, x)) for x in float_scores]
        int_scores = [round(max(0, min(score, 5))) for score in logits]
        pred_labels = ["high_quality" if score >= 2.5 else "low_quality" for score in logits]  # noqa: PLR2004

        return {
            self.float_score_column: float_scores,
            self.int_score_column: int_scores,
            self.pred_column: pred_labels,
        }

    def create_output_dataframe(self, df_cpu: pd.DataFrame, collected_output: dict[str, np.ndarray]) -> pd.DataFrame:
        df_cpu = df_cpu.drop(columns=[INPUT_ID_COLUMN, ATTENTION_MASK_COLUMN])

        df_cpu[self.float_score_column] = collected_output[self.float_score_column]
        df_cpu[self.int_score_column] = collected_output[self.int_score_column]
        df_cpu[self.pred_column] = collected_output[self.pred_column]

        return df_cpu


@dataclass(kw_only=True)
class _FineWebBaseClassifier(CompositeStage[DocumentBatch, DocumentBatch]):
    """
    Parent class for FineWebEduClassifier, FineWebMixtralEduClassifier, and FineWebNemotronEduClassifier,
    since their implementations are almost identical.

    Args:
        model_identifier: The identifier of the Hugging Face model.
        cache_dir: The Hugging Face cache directory. Defaults to None.
        pred_column: The name of the prediction column.
        float_score_column: The name of the float score column.
        int_score_column: The name of the integer score column.
        text_field: The name of the text field in the input data. Defaults to "text".
        filter_by: For categorical classifiers, the list of labels to filter the data by. Defaults to None.
        max_chars: Limits the total number of characters that can be fed to the tokenizer.
            If None, text will not be truncated. Defaults to None.
        max_seq_length: Limits the total sequence returned by the tokenizer so that it has a maximum length.
            Defaults to 512.
        sort_by_length: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        model_inference_batch_size: The size of the batch for model inference. Defaults to 256.
        autocast: Whether to use autocast. When True, we trade off minor accuracy for faster inference.
            Defaults to True.

    """

    model_identifier: str
    cache_dir: str | None = None
    pred_column: str = "preds"
    float_score_column: str = "float_score"
    int_score_column: str = "int_score"
    text_field: str = "text"
    filter_by: list[str] | None = None
    max_chars: int | None = None
    max_seq_length: int = MAX_SEQ_LENGTH
    sort_by_length: bool = True
    model_inference_batch_size: int = 256
    autocast: bool = True

    def __post_init__(self) -> None:
        super().__init__()

        self.stages = [
            TokenizerStage(
                model_identifier=self.model_identifier,
                cache_dir=self.cache_dir,
                text_field=self.text_field,
                max_chars=self.max_chars,
                max_seq_length=self.max_seq_length,
                padding_side=DEBERTA_TOKENIZER_PADDING_SIDE,
                sort_by_length=self.sort_by_length,
            ),
            FineWebModelStage(
                model_identifier=self.model_identifier,
                cache_dir=self.cache_dir,
                pred_column=self.pred_column,
                float_score_column=self.float_score_column,
                int_score_column=self.int_score_column,
                model_inference_batch_size=self.model_inference_batch_size,
                has_seq_order=self.sort_by_length,
                autocast=self.autocast,
            ),
        ]

        if self.filter_by is not None and len(self.filter_by) > 0:
            self.stages.append(Filter(filter_fn=self.filter_by_category, filter_field=self.pred_column))

    def inputs(self) -> tuple[list[str], list[str]]:
        return self.stages[0].inputs()

    def outputs(self) -> tuple[list[str], list[str]]:
        return self.stages[1].outputs()

    def filter_by_category(self, value: str) -> bool:
        return value in self.filter_by

    def decompose(self) -> list[ProcessingStage]:
        return self.stages


class FineWebEduClassifier(_FineWebBaseClassifier):
    """
    FineWebEduClassifier is a specialized classifier designed for educational content assessment,
    utilizing the Hugging Face FineWeb EDU Classifier model (https://huggingface.co/HuggingFaceFW/fineweb-edu-classifier).
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large text datasets.

    Attributes:
        cache_dir: The Hugging Face cache directory. Defaults to None.
        pred_column: The name of the prediction column. Defaults to "fineweb-edu-score-label".
        float_score_column: The name of the float score column. Defaults to "fineweb-edu-score-float".
        int_score_column: The name of the integer score column. Defaults to "fineweb-edu-score-int".
        text_field: The name of the text field in the input data. Defaults to "text".
        filter_by: For categorical classifiers, the list of labels to filter the data by. Defaults to None.
        max_chars: Limits the total number of characters that can be fed to the tokenizer.
            If None, text will not be truncated. Defaults to None.
        sort_by_length: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        model_inference_batch_size: The size of the batch for model inference. Defaults to 256.
        autocast: Whether to use autocast. When True, we trade off minor accuracy for faster inference.
            Defaults to True.

    """

    def __init__(  # noqa: PLR0913
        self,
        cache_dir: str | None = None,
        pred_column: str = "fineweb-edu-score-label",
        float_score_column: str = "fineweb-edu-score-float",
        int_score_column: str = "fineweb-edu-score-int",
        text_field: str = "text",
        filter_by: list[str] | None = None,
        max_chars: int | None = None,
        sort_by_length: bool = True,
        model_inference_batch_size: int = 256,
        autocast: bool = True,
    ):
        super().__init__(
            model_identifier=FINEWEB_EDU_MODEL_IDENTIFIER,
            cache_dir=cache_dir,
            pred_column=pred_column,
            float_score_column=float_score_column,
            int_score_column=int_score_column,
            text_field=text_field,
            filter_by=filter_by,
            max_chars=max_chars,
            max_seq_length=MAX_SEQ_LENGTH,
            sort_by_length=sort_by_length,
            model_inference_batch_size=model_inference_batch_size,
            autocast=autocast,
        )

        self._name = format_name_with_suffix(FINEWEB_EDU_MODEL_IDENTIFIER)


class FineWebMixtralEduClassifier(_FineWebBaseClassifier):
    """
    FineWebMixtralEduClassifier is a specialized classifier designed for educational content assessment,
    utilizing the NemoCurator FineWeb Mixtral Edu Classifier model (https://huggingface.co/nvidia/nemocurator-fineweb-mixtral-edu-classifier).
    It is similar to the FineWeb-Edu classifier and was trained on the same text samples, but using annotations from Mixtral 8x22B-Instruct.
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large text datasets.

    Attributes:
        cache_dir: The Hugging Face cache directory. Defaults to None.
        pred_column: The name of the prediction column. Defaults to "fineweb-mixtral-edu-score-label".
        float_score_column: The name of the float score column. Defaults to "fineweb-mixtral-edu-score-float".
        int_score_column: The name of the integer score column. Defaults to "fineweb-mixtral-edu-score-int".
        text_field: The name of the text field in the input data. Defaults to "text".
        filter_by: For categorical classifiers, the list of labels to filter the data by. Defaults to None.
        max_chars: Limits the total number of characters that can be fed to the tokenizer.
            If None, text will not be truncated. Defaults to None.
        sort_by_length: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        model_inference_batch_size: The size of the batch for model inference. Defaults to 1024.
        autocast: Whether to use autocast. When True, we trade off minor accuracy for faster inference.
            Defaults to True.

    """

    def __init__(  # noqa: PLR0913
        self,
        cache_dir: str | None = None,
        pred_column: str = "fineweb-mixtral-edu-score-label",
        float_score_column: str = "fineweb-mixtral-edu-score-float",
        int_score_column: str = "fineweb-mixtral-edu-score-int",
        text_field: str = "text",
        filter_by: list[str] | None = None,
        max_chars: int | None = None,
        sort_by_length: bool = True,
        model_inference_batch_size: int = 1024,
        autocast: bool = True,
    ):
        super().__init__(
            model_identifier=FINEWEB_MIXTRAL_EDU_MODEL_IDENTIFIER,
            cache_dir=cache_dir,
            pred_column=pred_column,
            float_score_column=float_score_column,
            int_score_column=int_score_column,
            text_field=text_field,
            filter_by=filter_by,
            max_chars=max_chars,
            max_seq_length=MAX_SEQ_LENGTH,
            sort_by_length=sort_by_length,
            model_inference_batch_size=model_inference_batch_size,
            autocast=autocast,
        )

        self._name = format_name_with_suffix(FINEWEB_MIXTRAL_EDU_MODEL_IDENTIFIER)


class FineWebNemotronEduClassifier(_FineWebBaseClassifier):
    """
    FineWebNemotronEduClassifier is a specialized classifier designed for educational content assessment,
    utilizing the NemoCurator FineWeb Nemotron-4 Edu Classifier model (https://huggingface.co/nvidia/nemocurator-fineweb-nemotron-4-edu-classifier).
    It is similar to the FineWeb-Edu classifier and was trained on the same text samples, but using annotations from Nemotron-4-340B-Instruct.
    This classifier is optimized for running on multi-node, multi-GPU setups to enable fast and efficient inference on large text datasets.

    Attributes:
        cache_dir: The Hugging Face cache directory. Defaults to None.
        pred_column: The name of the prediction column. Defaults to "fineweb-nemotron-edu-score-label".
        float_score_column: The name of the float score column. Defaults to "fineweb-nemotron-edu-score-float".
        int_score_column: The name of the integer score column. Defaults to "fineweb-nemotron-edu-score-int".
        text_field: The name of the text field in the input data. Defaults to "text".
        filter_by: For categorical classifiers, the list of labels to filter the data by. Defaults to None.
        max_chars: Limits the total number of characters that can be fed to the tokenizer.
            If None, text will not be truncated. Defaults to None.
        sort_by_length: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        model_inference_batch_size: The size of the batch for model inference. Defaults to 1024.
        autocast: Whether to use autocast. When True, we trade off minor accuracy for faster inference.
            Defaults to True.

    """

    def __init__(  # noqa: PLR0913
        self,
        cache_dir: str | None = None,
        pred_column: str = "fineweb-nemotron-edu-score-label",
        float_score_column: str = "fineweb-nemotron-edu-score-float",
        int_score_column: str = "fineweb-nemotron-edu-score-int",
        text_field: str = "text",
        filter_by: list[str] | None = None,
        max_chars: int | None = None,
        sort_by_length: bool = True,
        model_inference_batch_size: int = 1024,
        autocast: bool = True,
    ):
        super().__init__(
            model_identifier=FINEWEB_NEMOTRON_EDU_MODEL_IDENTIFIER,
            cache_dir=cache_dir,
            pred_column=pred_column,
            float_score_column=float_score_column,
            int_score_column=int_score_column,
            text_field=text_field,
            filter_by=filter_by,
            max_chars=max_chars,
            max_seq_length=MAX_SEQ_LENGTH,
            sort_by_length=sort_by_length,
            model_inference_batch_size=model_inference_batch_size,
            autocast=autocast,
        )

        self._name = format_name_with_suffix(FINEWEB_NEMOTRON_EDU_MODEL_IDENTIFIER)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import io
import json
import pathlib
import uuid
from typing import Any

import pandas as pd
from loguru import logger


class JsonEncoderCustom(json.JSONEncoder):
    """Custom JSON encoder that handles types that are not JSON serializable.

    Example:
    ```python
    json.dumps(data, cls=JsonEncoderClass)
    ```

    """

    def default(self, obj: object) -> str | object:
        """Encode an object for JSON serialization.

        Args:
            obj: Object to encode.

        Returns:
            Encoded object.

        """
        if isinstance(obj, uuid.UUID):
            return str(obj)
        return super().default(obj)  # type: ignore[no-any-return]


def write_bytes(  # noqa: PLR0913
    buffer: bytes,
    dest: pathlib.Path,
    desc: str,
    source_video: str,
    *,
    verbose: bool,
    backup_and_overwrite: bool = False,
    overwrite: bool = False,
) -> None:
    """Write bytes to local path.

    Args:
        buffer: Bytes to write.
        dest: Destination to write.
        desc: Description of the write.
        source_video: Source video.
        verbose: Verbosity.
        client: Storage client.
        backup_and_overwrite: Backup and overwrite.
        overwrite: Overwrite.

    """
    if dest.exists():
        if backup_and_overwrite:
            msg = "Backup and overwrite is not implemented"
            raise NotImplementedError(msg)
        elif overwrite:
            logger.warning(f"{desc} {dest} already exists, overwriting ...")
        else:
            logger.warning(f"{desc} {dest} already exists, skipping ...")
            return
    if verbose:
        logger.info(f"Writing {desc} for {source_video} to {dest}")
    dest.parent.mkdir(parents=True, exist_ok=True)
    with dest.open("wb") as fp:
        fp.write(buffer)


def write_parquet(  # noqa: PLR0913
    data: list[dict[str, str]],
    dest: pathlib.Path,
    desc: str,
    source_video: str,
    *,
    verbose: bool,
    backup_and_overwrite: bool = False,
    overwrite: bool = False,
) -> None:
    """Write parquet to local path.

    Args:
        data: Data to write.
        dest: Destination to write.
        desc: Description of the write.
        source_video: Source video.
        verbose: Verbosity.
        client: Storage client.
        backup_and_overwrite: Whether to backup existing file before overwriting.
        overwrite: Whether to overwrite existing file.

    """
    # Convert list of dicts to DataFrame
    pdf = pd.DataFrame(data)

    # Write DataFrame to a Parquet file in memory
    parquet_buffer = io.BytesIO()
    pdf.to_parquet(parquet_buffer, index=False)

    write_bytes(
        parquet_buffer.getvalue(),
        dest,
        desc,
        source_video,
        verbose=verbose,
        backup_and_overwrite=backup_and_overwrite,
        overwrite=overwrite,
    )


def write_json(  # noqa: PLR0913
    data: dict[str, Any],
    dest: pathlib.Path,
    desc: str,
    source_video: str,
    *,
    verbose: bool,
    backup_and_overwrite: bool = False,
    overwrite: bool = False,
) -> None:
    """Write json to local path.

    Args:
        data: Data to write.
        dest: Destination to write.
        desc: Description of the write.
        source_video: Source video.
        verbose: Verbosity.
        client: Storage client.
        backup_and_overwrite: Backup and overwrite.
        overwrite: Overwrite.

    """
    updated_json = json.dumps(data, indent=4, cls=JsonEncoderCustom)
    json_bytes = io.BytesIO(updated_json.encode("utf-8"))
    write_bytes(
        json_bytes.getvalue(),
        dest,
        desc,
        source_video,
        verbose=verbose,
        backup_and_overwrite=backup_and_overwrite,
        overwrite=overwrite,
    )


def write_csv(  # noqa: PLR0913
    dest: pathlib.Path,
    desc: str,
    source_video: str,
    data: list[list[str]],
    *,
    verbose: bool,
    backup_and_overwrite: bool = False,
) -> None:
    """Write csv to local path.

    Args:
        dest: Destination to write.
        desc: Description of the write.
        source_video: Source video.
        data: Data to write.
        verbose: Verbosity.
        client: Storage client.
        backup_and_overwrite: Backup and overwrite.

    """
    output = io.StringIO()
    writer = csv.writer(output)
    writer.writerows(data)
    csv_bytes = io.BytesIO(output.getvalue().encode("utf-8"))
    write_bytes(
        csv_bytes.getvalue(),
        dest,
        desc,
        source_video,
        verbose=verbose,
        backup_and_overwrite=backup_and_overwrite,
    )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any, Literal

from nemo_curator.backends.experimental.utils import RayStageSpecKeys
from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.deduplication.fuzzy.lsh.lsh import LSHActor
from nemo_curator.stages.deduplication.fuzzy.utils import CURATOR_DEFAULT_MINHASH_FIELD
from nemo_curator.stages.deduplication.id_generator import CURATOR_DEDUP_ID_STR
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks import FileGroupTask
from nemo_curator.utils.file_utils import create_or_overwrite_dir, get_fs


@dataclass
class LSHStage(ProcessingStage[FileGroupTask, FileGroupTask]):
    """
    Stage that performs LSH on a FileGroupTask containing minhash data.

    The executor will process this stage in iterations based on bands_per_iteration.

    Parameters
    ----------
    num_bands
        Number of LSH bands.
    minhashes_per_band
        Number of minhashes per band.
    id_field
        Name of the ID field in input data.
    minhash_field
        Name of the minhash field in input data.
    output_path
        Base path to write output files.
    read_kwargs
        Keyword arguments for the read method.
    write_kwargs
        Keyword arguments for the write method.
    rmm_pool_size
        Size of the RMM GPU memory pool in bytes.
        If "auto", the memory pool is set to 90% of the free GPU memory.
        If None, the memory pool is set to 50% of the free GPU memory that can expand if needed.
    spill_memory_limit
        Device memory limit in bytes for spilling to host.
        If "auto", the limit is set to 80% of the RMM pool size.
        If None spilling is disabled.
    enable_statistics
        Whether to collect statistics.
    bands_per_iteration
        Number of bands to process per shuffle iteration. Between 1 and num_bands.
        Higher values reduce the number of shuffle iterations but increase the memory usage.
    total_nparts
        Total number of partitions to write during the shuffle.
        If None, the number of partitions will be decided automatically by the executor as the closest power of 2 <= number of input tasks.
    """

    _name = "LSHStage"
    _resources = Resources(gpus=1.0)

    # Core Algo objects
    actor_class = LSHActor

    # LSH parameters
    num_bands: int
    minhashes_per_band: int
    # Data parameters
    id_field: str = CURATOR_DEDUP_ID_STR
    minhash_field: str = CURATOR_DEFAULT_MINHASH_FIELD
    output_path: str = "./"
    read_kwargs: dict[str, Any] | None = None
    write_kwargs: dict[str, Any] | None = None
    # Shuffle parameters
    rmm_pool_size: int | Literal["auto"] | None = "auto"
    spill_memory_limit: int | Literal["auto"] | None = "auto"
    enable_statistics: bool = False
    bands_per_iteration: int = 5  # number of bands to process in each iteration
    total_nparts: int | None = None

    def __post_init__(self):
        super().__init__()

        self.read_kwargs = self.read_kwargs if self.read_kwargs is not None else {}
        self.write_kwargs = self.write_kwargs if self.write_kwargs is not None else {}
        self.output_paths = []

        self.actor_kwargs = {
            "num_bands": self.num_bands,
            "minhashes_per_band": self.minhashes_per_band,
            "id_field": self.id_field,
            "minhash_field": self.minhash_field,
            "rmm_pool_size": self.rmm_pool_size,
            "spill_memory_limit": self.spill_memory_limit,
            "enable_statistics": self.enable_statistics,
            "read_kwargs": self.read_kwargs,
            "write_kwargs": self.write_kwargs,
            "total_nparts": self.total_nparts,  # Can be None, executor will set it
        }

        if self.bands_per_iteration < 1 or self.bands_per_iteration > self.num_bands:
            err_msg = (
                f"Invalid bands_per_iteration: {self.bands_per_iteration}, must be in range [1, {self.num_bands}]"
            )
            raise ValueError(err_msg)

        # Handle output directory and subdirectories
        output_fs = get_fs(self.output_path, storage_options=self.write_kwargs.get("storage_options"))
        output_base_path = output_fs.sep.join([self.output_path, self.name])

        create_or_overwrite_dir(output_base_path, fs=output_fs)

        for band_range in self.get_band_iterations():
            output_path = output_fs.sep.join([output_base_path, f"band_{band_range[0]}-band_{band_range[1]}"])
            create_or_overwrite_dir(output_path, fs=output_fs)
            self.output_paths.append(output_path)

    def process(self, task: FileGroupTask) -> FileGroupTask:
        err_msg = "LSHProcessingStage does not support the process method."
        raise NotImplementedError(err_msg)

    def ray_stage_spec(self) -> dict[str, Any]:
        """Ray stage specification for this stage."""
        return {
            RayStageSpecKeys.IS_LSH_STAGE: True,
        }

    def _check_actor_obj(self) -> None:
        if not hasattr(self, "_actor_obj") or not isinstance(self._actor_obj, self.actor_class):
            error = "Actor object not initialized. This might be because an incorrect executor was used or it failed to setup the stage properly."
            raise RuntimeError(error)

    def read_and_insert(self, task: FileGroupTask, band_range: tuple[int, int]) -> FileGroupTask:
        self._check_actor_obj()
        result = self._actor_obj.read_and_insert(task.data, band_range)
        self._current_band_range = band_range
        self.output_columns = result
        self.dataset_name = task.dataset_name
        return task

    def insert_finished(self) -> None:
        self._check_actor_obj()
        self._actor_obj.insert_finished()

    def extract_and_write(self) -> list[FileGroupTask]:
        self._check_actor_obj()
        current_band_min, current_band_max = self._current_band_range
        partition_dicts = self._actor_obj.extract_and_write()
        return [
            FileGroupTask(
                task_id=f"b{current_band_min}_b{current_band_max}_{partition_info['partition_id']}",
                dataset_name=self.dataset_name + f"{self.name}",
                data=[partition_info["path"]],
                _metadata={
                    "partition_index": partition_info["partition_id"],
                    "total_partitions": len(partition_dicts),
                    "num_docs": partition_info["num_docs"],
                    "output_columns": self.output_columns,
                },
            )
            for partition_info in partition_dicts
        ]

    def teardown(self) -> None:
        self._check_actor_obj()
        self._actor_obj.cleanup()

    def get_band_iterations(self) -> Iterator[tuple[int, int]]:
        """Get all band ranges for iteration."""
        for band_start in range(0, self.num_bands, self.bands_per_iteration):
            band_range = (band_start, min(band_start + self.bands_per_iteration, self.num_bands))
            yield band_range



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib


def _get_local_path(localpath: pathlib.Path, *args: str) -> pathlib.Path:
    """Construct a full local path from a base path and additional components.
    Args:
        localpath: The base local path.
        *args: Additional path components.
    Returns:
        The full local path as a Path object.
    """
    return pathlib.Path(localpath, *args)


def get_full_path(path: str | pathlib.Path, *args: str) -> pathlib.Path:
    """Construct a full path from a base path and additional components.
    Args:
        path: The base path.
        *args: Additional path components.
    Returns:
        The full path as a StoragePrefix or Path object.
    """
    # Convert string paths to the appropriate type
    if isinstance(path, str):
        return _get_local_path(pathlib.Path(path), *args)

    return pathlib.Path(path, *args)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Any

from nemo_curator.stages.text.download import DocumentExtractor

# The iterator and extractor code are in large part taken
# from the Red-Pajama repo
# https://github.com/togethercomputer/RedPajama-Data/tree/main/data_prep/arxiv


class ArxivExtractor(DocumentExtractor):
    """Extracts text from Arxiv LaTeX files."""

    def __init__(self):
        super().__init__()

    def _build_non_arg_macros_dict(self, file_content: str) -> dict[str, str]:
        r"""function takes the content of a tex file and returns a dictionary
        that contains the definitions of all macros that do not use arguments.
        The dictionary is of the form {macro_name: macro_value}.

        @param file_content: the content of the tex file as a string.

        @return: dict
        """
        # regex for extracting \newcommand macros without arguments
        non_arg_nc_reg = re.compile(
            # this regex matches the following:
            # \newcommand{\macro_name}{macro_value}
            # \newcommand*{\macro_name}{macro_value}
            # where macro_name is only allowed to contain letters and numbers;
            # macro_value can contain any character.
            pattern=r"\\\bnewcommand\b\*?\{(\\[a-zA-Z0-9]+?)\}\{(.*?)\}$",
            flags=re.MULTILINE,
        )

        # regex for extracting \def macros without arguments
        non_arg_def_reg = re.compile(
            # this regex matches the following:
            # \def\macro_name{macro_value}
            # where macro_name is only allowed to contain letters and numbers;
            # macro_value can contain any character.
            pattern=r"\\def\s*(\\[a-zA-Z0-9]+?)\s*\{(.*?)\}$",
            flags=re.MULTILINE,
        )

        # Extract all user-defined LaTeX macros from the preamble
        macros = {}
        for reg in [non_arg_nc_reg, non_arg_def_reg]:
            for match in reg.finditer(file_content):
                # convert the macro name and value to a raw string that can be
                # used in re.sub
                macro_name = match.group(1).encode("unicode-escape").decode("utf-8")
                macro_val = match.group(2).encode("unicode-escape").decode("utf-8")

                macros[macro_name] = macro_val

        return macros

    def _clean_tex_file(self, file_content: str, arg_macros: dict[str, str], non_arg_macros: dict[str, str]) -> str:
        r"""function takes a tex file as input and returns a cleaned version. The
         cleaned version is a concatenation of the tex files with the
        following modifications:

        - remove all comments (i.e. all lines starting with %)
        - remove everything before the first section-like header
        - remove everything after the first occurrence of either \appendix or
            \bibliography
        - inline-expand definitions and macros

        @param file_content: the content of the tex file as a string.

        @return: cleaned tex file as a string
        """
        # find the first occurence of a \section-like header and replace everything
        # before it with an empty string. This matches the following pattern:
        #   \<section-type>[optional-args]{name}
        pattern = r"^(.*?)("
        pattern += r"\\\bchapter\b\*?(?:\[(.*?)\])?\{(.*?)\}|"
        pattern += r"\\\bpart\b\*?(?:\[(.*?)\])?\{(.*?)\}|"
        pattern += r"\\\bsection\b\*?(?:\[(.*?)\])?\{(.*?)\}|"
        pattern += r"\\\bsubsection\b\*?(?:\[(.*?)\])?\{(.*?)\}|"
        pattern += r"\\\bsubsubsection\b\*?(?:\[(.*?)\])?\{(.*?)\}|"
        pattern += r"\\\bparagraph\b\*?(?:\[(.*?)\])?\{(.*?)\}"
        pattern += r"\\\bsubparagraph\b\*?(?:\[(.*?)\])?\{(.*?)\}"
        pattern += r")"

        # if no section like header is found, then we return an empty string
        if not re.search(pattern, file_content, flags=re.DOTALL):
            return ""

        # replace everything with the second group of the match (i.e. everything
        # after and including the section header)
        file_content = re.sub(
            pattern=pattern,
            repl=r"\2",
            string=file_content,
            flags=re.DOTALL,  # make sure that the dot matches also newlines
        )

        # remove all line comments
        file_content = re.sub(
            pattern=r"(?m)^%.*\n?",
            repl=r"",
            string=file_content,
            flags=re.MULTILINE,
        )

        # remove all in comments within a line
        file_content = re.sub(
            # pattern matches a "%" that is not preceded by a backslash (=comment)
            pattern=r"[^\\]%.+$",
            repl=r"",
            string=file_content,
            flags=re.MULTILINE,
        )

        # find the first occurence of either \appendix or \bibliography and
        # replace everything after it with an empty string
        pattern = r"("
        pattern += r"\\appendix|"
        pattern += r"\\begin\{references\}|"
        pattern += r"\\begin\{REFERENCES\}|"
        pattern += r"\\begin\{thebibliography\}|"
        pattern += r"\\bibliography\{.*\}"
        pattern += r").*$"

        file_content = re.sub(
            pattern=pattern,
            repl=r"",
            string=file_content,
            flags=re.DOTALL,  # make sure that the dot matches also newlines
        )

        # inline-expand all non-arg macros
        for macro_name, macro_value in non_arg_macros.items():
            file_content = re.sub(
                # make pattern grouped to make sure that the macro is not part
                # of a longer alphanumeric word
                pattern=r"(" + macro_name + r")" + r"([^a-zA-Z0-9])",
                # replace the macro with its value and add back the character that
                # was matched after the macro
                repl=macro_value + r"\2",
                string=file_content,
            )

        # inline-expand all macros that use args
        # TODO: inline-expand macros with args
        for _macro_name, _macro_value in arg_macros.items():
            pass

        return file_content

    def extract(self, record: dict[str, str]) -> dict[str, Any] | None:
        if "content" not in record or len(record["content"]) == 0:
            return None

        # build dictionaries that contain the definitions of all macros in all tex
        # files. This is later used to expand all macros used in the text with
        # their definitions, so that consistency among different authors is
        # ensured.

        non_arg_macros = {}
        for file_content in record["content"]:
            non_arg_macros.update(self._build_non_arg_macros_dict(file_content))

        # TODO: macros that take arguments are not supported yet
        arg_macros = {}

        # join multiple latex files with a newline character
        try:
            cleaned_latex_file_str = "\n".join(
                self._clean_tex_file(
                    file_content=file_content,
                    arg_macros=arg_macros,
                    non_arg_macros=non_arg_macros,
                )
                for file_content in record["content"]
            )
        except Exception:  # noqa: BLE001
            return None

        # Don't return meta
        if (cleaned_latex_file_str is not None) and (len(cleaned_latex_file_str) > 0):
            return {"text": cleaned_latex_file_str}

        return None

    def input_columns(self) -> list[str]:
        return ["id", "source_id", "content"]

    def output_columns(self) -> list[str]:
        return ["text"]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import lru_cache
from typing import Any, Literal

os.environ["RAPIDS_NO_INITIALIZE"] = "1"

import numpy as np
import torch
from huggingface_hub import snapshot_download
from transformers import AutoConfig, AutoTokenizer

from nemo_curator.backends.base import NodeInfo, WorkerMetadata
from nemo_curator.stages.base import ProcessingStage
from nemo_curator.tasks import DocumentBatch

from .utils import (
    ATTENTION_MASK_COLUMN,
    INPUT_ID_COLUMN,
    SEQ_ORDER_COLUMN,
    TOKEN_LENGTH_COLUMN,
    format_name_with_suffix,
)


class TokenizerStage(ProcessingStage[DocumentBatch, DocumentBatch]):
    """
    Tokenizer stage for Hugging Face models.

    Args:
        model_identifier: The identifier of the Hugging Face model.
        cache_dir: The Hugging Face cache directory. Defaults to None.
        hf_token: Hugging Face token for downloading the model, if needed. Defaults to None.
        text_field: The name of the text field in the input data. Defaults to "text".
        max_chars: Limits the total number of characters that can be fed to the tokenizer.
            If None, text will not be truncated. Defaults to None.
        max_seq_length: Limits the total sequence returned by the tokenizer so that it has a maximum length.
            If None, the tokenizer's model_max_length is used. Defaults to None.
        padding_side: The side to pad the input tokens. Defaults to "right".
        sort_by_length: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        unk_token: If True, set the pad_token to the tokenizer's unk_token. Defaults to False.

    """

    def __init__(  # noqa: PLR0913
        self,
        model_identifier: str,
        cache_dir: str | None = None,
        hf_token: str | None = None,
        text_field: str = "text",
        max_chars: int | None = None,
        max_seq_length: int | None = None,
        padding_side: Literal["left", "right"] = "right",
        sort_by_length: bool = True,
        unk_token: bool = False,
    ):
        self._name = format_name_with_suffix(model_identifier, suffix="_tokenizer")

        self.model_identifier = model_identifier
        self.cache_dir = cache_dir
        self.hf_token = hf_token
        self.text_field = text_field
        self.max_chars = max_chars
        self.max_seq_length = max_seq_length
        self.padding_side = padding_side
        self.sort_by_length = sort_by_length
        self.unk_token = unk_token

    def inputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [self.text_field]

    def outputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [self.text_field, INPUT_ID_COLUMN, ATTENTION_MASK_COLUMN] + (
            [SEQ_ORDER_COLUMN] if self.sort_by_length else []
        )

    def ray_stage_spec(self) -> dict[str, Any]:
        return {"is_actor_stage": True}

    def setup_on_node(self, _node_info: NodeInfo | None = None, _worker_metadata: WorkerMetadata = None) -> None:
        try:
            snapshot_download(
                repo_id=self.model_identifier,
                cache_dir=self.cache_dir,
                token=self.hf_token,
                local_files_only=False,
            )
            self._setup(local_files_only=False)
        except Exception as e:
            msg = f"Failed to download {self.model_identifier}"
            raise RuntimeError(msg) from e

    @lru_cache(maxsize=1)  # noqa: B019
    def load_cfg(self, local_files_only: bool = True) -> AutoConfig:
        return AutoConfig.from_pretrained(
            self.model_identifier, cache_dir=self.cache_dir, local_files_only=local_files_only
        )

    # We use the _setup function to ensure that everything needed for the tokenizer is downloaded and loaded properly
    def _setup(self, local_files_only: bool = True) -> None:
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_identifier,
            padding_side=self.padding_side,
            cache_dir=self.cache_dir,
            local_files_only=local_files_only,
        )
        if self.unk_token:
            self.tokenizer.pad_token = self.tokenizer.unk_token

        if self.max_seq_length is None:
            self.max_seq_length = self.tokenizer.model_max_length

            # Guard against the HF bug
            # which sets max_seq_length to max(int) for some models
            if self.max_seq_length > 1e5:  # noqa: PLR2004
                self.max_seq_length = self.load_cfg(local_files_only=local_files_only).max_position_embeddings

    def setup(self, _: WorkerMetadata | None = None) -> None:
        self._setup(local_files_only=True)

    def process(self, batch: DocumentBatch) -> DocumentBatch:
        df = batch.to_pandas()

        if self.max_chars is not None and self.max_chars > 0:
            df[self.text_field] = df[self.text_field].str.slice(0, self.max_chars)

        with torch.no_grad():
            tokens = self.tokenizer.batch_encode_plus(
                df[self.text_field].tolist(),
                max_length=self.max_seq_length,
                padding="max_length",
                return_tensors="np",
                truncation=True,
                add_special_tokens=True,
                return_token_type_ids=False,
            )

        output = df.copy()
        output[INPUT_ID_COLUMN] = tokens.input_ids.tolist()
        output[ATTENTION_MASK_COLUMN] = tokens.attention_mask.tolist()

        if self.sort_by_length:
            # Add column to preserve original order
            output[SEQ_ORDER_COLUMN] = np.arange(len(df))
            output[TOKEN_LENGTH_COLUMN] = tokens.attention_mask.sum(axis=1)
            output = output.sort_values(by=TOKEN_LENGTH_COLUMN, kind="stable", ignore_index=True).drop(
                columns=[TOKEN_LENGTH_COLUMN]
            )

        return DocumentBatch(
            task_id=batch.task_id,
            dataset_name=batch.dataset_name,
            data=output,
            _metadata=batch._metadata,
            _stage_perf=batch._stage_perf,
        )



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from nemo_curator.tasks import Task
from nemo_curator.utils.performance_utils import StageTimer

if TYPE_CHECKING:
    from nemo_curator.stages.base import ProcessingStage


@dataclass
class NodeInfo:
    """Generic node information for setup_on_node calls across backends.
    Simplified to match Xenna's structure.
    """

    node_id: str = ""


@dataclass
class WorkerMetadata:
    """Generic worker metadata for setup_on_node calls across backends.
    Simplified to match Xenna's structure. The allocation field can contain
    backend-specific allocation information.
    """

    worker_id: str = ""
    allocation: Any = None  # Backend-specific allocation info


class BaseExecutor(ABC):
    """Executor for a pipeline."""

    def __init__(self, config: dict[str, Any] | None = None):
        self.config = config or {}

    @abstractmethod
    def execute(self, stages: list["ProcessingStage"], initial_tasks: list[Task] | None = None) -> None:
        """Execute the pipeline."""


class BaseStageAdapter:
    """Adapts ProcessingStage to an execution backend, if needed."""

    def __init__(self, stage: "ProcessingStage"):
        self.stage = stage

    def process_batch(self, tasks: list[Task]) -> list[Task]:
        """Process a batch of tasks.

        Args:
            tasks (list[Task]): List of tasks to process

        Returns:
            list[Task]: List of processed tasks
        """
        # Lazy initialize timer if needed
        if not hasattr(self, "_timer") or self._timer is None:
            self._timer = StageTimer(self.stage)

        # Calculate input data size for timer
        input_size = sum(task.num_items for task in tasks)
        # Initialize performance timer for this batch
        self._timer.reinit(input_size)

        with self._timer.time_process(input_size):
            # Use the batch processing logic
            results = self.stage.process_batch(tasks)

        # Log performance stats and add to result tasks
        _, stage_perf_stats = self._timer.log_stats()
        # Consume and attach any custom metrics recorded by the stage during this call
        custom_metrics = self.stage._consume_custom_metrics()
        if custom_metrics:
            stage_perf_stats.custom_metrics.update(custom_metrics)
        for task in results:
            task.add_stage_perf(stage_perf_stats)

        return results

    def setup_on_node(self, node_info: NodeInfo | None = None, worker_metadata: WorkerMetadata | None = None) -> None:
        """Setup the stage on a node.

        Args:
            node_info (NodeInfo, optional): Information about the node
            worker_metadata (WorkerMetadata, optional): Information about the worker
        """
        # Call the underlying stage's setup_on_node method
        # Some backends may provide node/worker info, others may not
        self.stage.setup_on_node(node_info, worker_metadata)

    def setup(self, worker_metadata: WorkerMetadata | None = None) -> None:
        """Setup the stage once per actor.

        Args:
            worker_metadata (WorkerMetadata, optional): Information about the worker
        """
        self.stage.setup(worker_metadata)

    def teardown(self) -> None:
        """Teardown the stage once per actor."""
        self.stage.teardown()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Literal

import cudf
from loguru import logger
from rapidsmpf.utils.cudf import pylibcudf_to_cudf_dataframe

from nemo_curator.stages.deduplication.fuzzy.utils import CURATOR_DEFAULT_MINHASH_FIELD, CURATOR_LSH_BUCKET_FIELD
from nemo_curator.stages.deduplication.id_generator import CURATOR_DEDUP_ID_STR
from nemo_curator.stages.deduplication.shuffle_utils.rapidsmpf_shuffler import BulkRapidsMPFShuffler

if TYPE_CHECKING:
    from collections.abc import Iterator


class LSHActor(BulkRapidsMPFShuffler):
    """
    Actor that performs LSH operations and shuffling using Ray.

    Parameters
    ----------
    nranks
        Number of ranks in the communication group.
    total_nparts
        Total number of output partitions.
    num_bands
        Number of LSH bands.
    minhashes_per_band
        Number of minhashes per band.
    id_field
        Name of the ID field in input data.
    minhash_field
        Name of the minhash field in input data.
    output_path
        Path to write output files.
    rmm_pool_size
        Size of the RMM GPU memory pool in bytes.
        If "auto", the memory pool is set to 90% of the free GPU memory.
        If None, the memory pool is set to 50% of the free GPU memory that can expand if needed.
    spill_memory_limit
        Device memory limit in bytes for spilling to host.
        If "auto", the limit is set to 80% of the RMM pool size.
        If None spilling is disabled.
    enable_statistics
        Whether to collect statistics.
    read_kwargs
        Keyword arguments for the read method.
    write_kwargs
        Keyword arguments for the write method.

    Notes
    -----
    Architecture and Processing Flow:

    This implementation follows a clean separation of responsibilities with distinct methods
    for each part of the pipeline:

    Input Phase:
    - `read_minhash`: Reads minhash files and returns a DataFrame

    Processing Phase:
    - `minhash_to_bands`: Transforms a single minhash DataFrame into LSH bands
    - `read_and_insert`: Orchestrates reading, band creation, and insertion

    Output Phase:
    - `extract_and_group`: Extracts and groups shuffled data, yielding results as a generator
    - `extract_and_write`: Processes each yielded result and writes to output files immediately

    1. Files are read using `read_minhash`
    2. Data is processed with `minhash_to_bands` to extract LSH bucket IDs
    3. Processed data is immediately inserted into the shuffler
    4. Results are extracted and processed one partition at a time using generators
    5. Each partition is written to disk as soon as it's processed, without accumulating in memory
    """

    def __init__(  # noqa: PLR0913
        self,
        nranks: int,
        total_nparts: int,
        num_bands: int,
        minhashes_per_band: int,
        id_field: str = CURATOR_DEDUP_ID_STR,
        minhash_field: str = CURATOR_DEFAULT_MINHASH_FIELD,
        output_path: str = "./",
        rmm_pool_size: int | Literal["auto"] | None = "auto",
        spill_memory_limit: int | Literal["auto"] | None = "auto",
        *,
        enable_statistics: bool = False,
        read_kwargs: dict[str, Any] | None = None,
        write_kwargs: dict[str, Any] | None = None,
    ):
        super().__init__(
            nranks=nranks,
            total_nparts=total_nparts,
            shuffle_on=[CURATOR_LSH_BUCKET_FIELD],
            output_path=output_path,
            rmm_pool_size=rmm_pool_size,
            spill_memory_limit=spill_memory_limit,
            enable_statistics=enable_statistics,
        )
        self.num_bands = num_bands
        self.minhashes_per_band = minhashes_per_band
        self.id_field = id_field
        self.minhash_field = minhash_field
        self.read_kwargs = read_kwargs if read_kwargs is not None else {}
        self.write_kwargs = write_kwargs if write_kwargs is not None else {}

    @staticmethod
    def _generate_band_ranges(num_bands: int, minhashes_per_band: int) -> list[list[int]]:
        """
        Generates a list of indices for the minhash ranges given num_bands &
        minhashes_per_band.
        eg: num_bands=3, minhashes_per_band=2
        [[0, 1], [2, 3], [4, 5]]
        """
        return [list(range(band * minhashes_per_band, (band + 1) * minhashes_per_band)) for band in range(num_bands)]

    def read_minhash(self, filepaths: list[str]) -> cudf.DataFrame:
        """
        Read minhash data from parquet files.

        Parameters
        ----------
        filepaths
            List of paths to minhash files.

        Returns
        -------
            DataFrame containing minhash data from all input files.
        """
        if self.read_kwargs.get("columns", None) is not None:
            err_msg = "Columns cannot be set in read_kwargs for LSHActor"
            raise ValueError(err_msg)
        return cudf.read_parquet(filepaths, columns=[self.id_field, self.minhash_field], **self.read_kwargs)

    def minhash_to_bands(self, minhash_df: cudf.DataFrame, band_range: tuple[int, int]) -> cudf.DataFrame:
        """
        Process a single minhash DataFrame to extract LSH band data.

        Parameters
        ----------
        minhash_df
            DataFrame containing minhash data.
        band_range
            Tuple of (start_band, end_band) to process.

        Returns
        -------
            DataFrame with document IDs and their corresponding bucket IDs.
        """
        if minhash_df is None or len(minhash_df) == 0:
            return None

        # Get the band ranges for the specified band range
        band_ranges = self._generate_band_ranges(num_bands=self.num_bands, minhashes_per_band=self.minhashes_per_band)[
            band_range[0] : band_range[1]
        ]

        id_df = minhash_df[[self.id_field]]

        for i, h in enumerate(band_ranges):
            indices = cudf.Series([h]).repeat(len(id_df))
            id_df[f"_bucket_{i}"] = f"b{i}_" + minhash_df[self.minhash_field].list.take(indices).hash_values(
                method="md5"
            )

        value_vars = [f"_bucket_{i}" for i in range(len(band_ranges))]
        melted_df = id_df.melt(id_vars=[self.id_field], value_name=CURATOR_LSH_BUCKET_FIELD, value_vars=value_vars)

        # Keep only the columns we need
        return melted_df[[self.id_field, CURATOR_LSH_BUCKET_FIELD]]

    def read_and_insert(self, filepaths: list[str], band_range: tuple[int, int]) -> None:
        """
        Read minhashes from files, create LSH bands, and insert into the shuffler.

        This method orchestrates the full processing pipeline:
        1. Reads minhash data from parquet files in batches
        2. Processes each batch to extract LSH bands
        3. Inserts the bands into the shuffler for distribution

        Parameters
        ----------
        filepaths
            List of paths to minhash files.
        band_range
            Tuple of (start_band, end_band) to process.

        Returns
        -------
        None
        """
        if not filepaths:
            return

        if band_range[0] < 0 or band_range[1] > self.num_bands or band_range[0] >= band_range[1]:
            msg = f"Invalid band range: {band_range}, must be in range [0, {self.num_bands}]"
            raise ValueError(msg)

        # Process files in batches
        minhash_df = self.read_minhash(filepaths)
        # Skip processing if the batch is empty
        if minhash_df is None or len(minhash_df) == 0:
            logger.info("Skipping empty batch")
            return

        # Process this batch of minhashes to get band data
        band_df = self.minhash_to_bands(minhash_df, band_range)

        # Call parent's insert_chunk method
        self.insert_chunk(band_df, list(band_df.columns))
        # Clear memory after processing a batch
        del minhash_df, band_df

    def group_by_bucket(self, df: cudf.DataFrame, include_singles: bool = False) -> cudf.DataFrame:
        """
        Group items by bucket ID and aggregate IDs into lists.

        Parameters
        ----------
        df
            DataFrame containing bucket IDs and document IDs.
        include_singles
            If True, include buckets with only one document. Default is False, which
            excludes single-document buckets as they cannot form duplicates. Set to True
            when building an LSH index that needs to maintain all documents.

        Returns
        -------
            DataFrame with bucket IDs and lists of document IDs.
        """
        if len(df) == 0:
            return df
        if not include_singles:
            # TODO: Add support for generating LSH index with single-document buckets that can be reused in incremental runs
            # Find bucket_ids that appear more than once (have multiple documents)
            # Keep only rows with buckets that are duplicated
            df = df[df[CURATOR_LSH_BUCKET_FIELD].duplicated(keep=False)]
        # Group by bucket_id and aggregate document IDs
        return df.groupby(CURATOR_LSH_BUCKET_FIELD)[self.id_field].agg(list).list.sort_values().reset_index()

    def extract_and_group(self) -> Iterator[tuple[int, cudf.DataFrame]]:
        """
        Extract shuffled partitions and group by bucket ID, yielding results one by one.

        This generator approach allows processing each partition immediately after it's ready,
        which is more memory-efficient than collecting all partitions first.

        Yields
        ------
        tuple
            A tuple of (partition_id, grouped_df) where grouped_df contains bucket IDs
            and their corresponding document ID lists.
        """
        # Fixed column names for pylibcudf conversion
        column_names = [self.id_field, CURATOR_LSH_BUCKET_FIELD]
        for partition_id, partition in self.extract():
            # Convert to cuDF DataFrame
            df = pylibcudf_to_cudf_dataframe(partition, column_names=column_names)
            # Group by bucket ID
            grouped_df = self.group_by_bucket(df)

            # Yield the result immediately instead of collecting in a list
            yield partition_id, grouped_df
            # Clean up memory
            del df, grouped_df

    def extract_and_write(self) -> list[dict[str, Any]]:
        """
        Extract shuffled partitions, group by bucket ID, and write results to files.

        This method orchestrates the post-processing pipeline:
        1. Extracts partitioned data from the shuffler using extract_and_group
        2. Writes each grouped partition to a parquet file as soon as it's available

        This generator-based approach is more memory-efficient since it processes
        one partition at a time rather than collecting all partitions in memory.

        Returns
        -------
        list[dict[str, Any]]
            A list of dictionaries containing partition information.
            Each dictionary contains:
            - partition_id: The ID of the partition
            - path: The path to the partition file
            - num_docs: The number of documents in the partition
        """
        partition_paths = []
        write_kwargs = self.write_kwargs.copy()
        write_kwargs["index"] = write_kwargs.get("index", False)
        # Process each partition as it becomes available
        for partition_id, grouped_df in self.extract_and_group():
            path = f"{self.output_path}/part.{partition_id}.parquet"

            # Write to file immediately
            grouped_df.to_parquet(path, **write_kwargs)
            partition_paths.append({"partition_id": partition_id, "path": path, "num_docs": len(grouped_df)})
            # Clean up to release memory
            del grouped_df

        return partition_paths



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class DocumentFilter(ABC):
    """
    An abstract base class for text-based document filters.

    This class serves as a template for creating specific document filters
    in the library. Subclasses should implement the abstract methods to
    define custom filtering behavior.
    """

    def __init__(self):
        super().__init__()
        self._name = self.__class__.__name__
        self._sentences = None
        self._paragraphs = None
        self._ngrams = None

    @abstractmethod
    def score_document(self, text: str) -> float | list[int | float]:
        """
        Calculate a score for the given document text.

        This method should be implemented by subclasses to define how
        a document's text is evaluated and scored.

        Args:
            text (str): The text content of the document to be scored.

        Returns:
            Any: A score or set of scores representing the document's
            relevance or quality. The type and structure of the
            return value should be consistent for each subclass.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """
        msg = "score_document method must be implemented by subclasses"
        raise NotImplementedError(msg)

    @abstractmethod
    def keep_document(self, scores: float | list[int | float]) -> bool:
        """
        Determine whether to keep a document based on its scores.

        This method should be implemented by subclasses to define the
        criteria for keeping or discarding a document based on the
        scores calculated by score_document().

        Args:
            scores (float | list[int | float]): The score or set of scores returned by score_document().
                          The type should match what is returned by score_document().

        Returns:
            bool: True if the document should be kept, False otherwise.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """
        msg = "keep_document method must be implemented by subclasses"
        raise NotImplementedError(msg)

    @property
    def name(self) -> str:
        return self._name

    @property
    def sentences(self) -> list:
        return self._sentences

    @sentences.setter
    def sentences(self, sentences: list) -> None:
        self._sentences = sentences

    @property
    def paragraphs(self) -> list:
        return self._paragraphs

    @paragraphs.setter
    def paragraphs(self, paragraphs: list) -> None:
        self._paragraphs = paragraphs

    @property
    def ngrams(self) -> dict:
        return self._ngrams

    @ngrams.setter
    def ngrams(self, ngrams: dict) -> None:
        self._ngrams = ngrams



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Model for fast shot transition detection.

@article{soucek2020transnetv2,
    title={TransNet V2: An effective deep network architecture for fast shot transition detection},
    author={Sou{\v{c}}ek, Tom{\'a}{\v{s}} and Loko{\v{c}}, Jakub},
    year={2020},
    journal={arXiv preprint arXiv:2008.04838},
}
"""

import pathlib
from collections.abc import Callable
from typing import Final

import torch
from loguru import logger
from torch import nn
from torch.nn import functional

from nemo_curator.utils.hf_download_utils import download_model_from_hf

from .base import ModelInterface

_TRANSNETV2_MODEL_ID: Final = "Sn4kehead/TransNetV2"
_TRANSNETV2_MODEL_WEIGHTS: Final = "transnetv2-pytorch-weights.pth"
_TRANSNETV2_MODEL_REVISION: Final = "db6ceab"

class _TransNetV2(nn.Module):
    def __init__(  # noqa: PLR0913
        self,
        rf: int = 16,
        rl: int = 3,
        rs: int = 2,
        rd: int = 1024,
        *,
        use_many_hot_targets: bool = True,
        use_frame_similarity: bool = True,
        use_color_histograms: bool = True,
        use_mean_pooling: bool = False,
        dropout_rate: float = 0.5,
    ) -> None:
        """Initialize the TransNetV2 model.

        Args:
            rf: Number of filters in the first layer.
            rl: Number of layers in the network.
            rs: Number of blocks in the network.
            rd: Number of output features.
            use_many_hot_targets: Whether to use many-hot targets.
            use_frame_similarity: Whether to use frame similarity.
            use_color_histograms: Whether to use color histograms.
            use_mean_pooling: Whether to use mean pooling.
            dropout_rate: Dropout rate.

        """
        super().__init__()
        self.SDDCNN = nn.ModuleList(
            [StackedDDCNNV2(in_filters=3, n_blocks=rs, filters=rf, stochastic_depth_drop_prob=0.0)]
            + [
                StackedDDCNNV2(in_filters=(rf * 2 ** (i - 1)) * 4, n_blocks=rs, filters=rf * 2**i)
                for i in range(1, rl)
            ],
        )

        self.frame_sim_layer = (
            FrameSimilarity(
                sum([(rf * 2**i) * 4 for i in range(rl)]),
                lookup_window=101,
                output_dim=128,
                similarity_dim=128,
                use_bias=True,
            )
            if use_frame_similarity
            else None
        )
        self.color_hist_layer = ColorHistograms(lookup_window=101, output_dim=128) if use_color_histograms else None

        self.dropout = nn.Dropout(dropout_rate) if dropout_rate is not None else None

        output_dim = ((rf * 2 ** (rl - 1)) * 4) * 3 * 6  # 3x6 for spatial dimensions
        if use_frame_similarity:
            output_dim += 128
        if use_color_histograms:
            output_dim += 128

        self.fc1 = nn.Linear(output_dim, rd)
        self.cls_layer1 = nn.Linear(rd, 1)
        self.cls_layer2 = nn.Linear(rd, 1) if use_many_hot_targets else None

        self.use_mean_pooling = use_mean_pooling
        self.eval()

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Process input through the TransNetV2 model.

        Args:
            inputs: Input tensor of video frames.

        Returns:
            Model predictions for shot transitions.

        """
        if not isinstance(inputs, torch.Tensor):
            msg = "inputs must be a torch.Tensor"
            raise TypeError(msg)
        if list(inputs.shape[2:]) != [27, 48, 3]:
            msg = f"incorrect shape: expected [*, *, 27, 48, 3], got {inputs.shape}"
            raise ValueError(msg)
        if inputs.dtype != torch.uint8:
            msg = f"incorrect dtype: expected torch.uint8, got {inputs.dtype}"
            raise ValueError(msg)
        # uint8 of shape [B, T, H, W, 3] to float of shape [B, 3, T, H, W]
        x = inputs.permute([0, 4, 1, 2, 3]).float()
        x = x.div_(255.0)

        block_features = []
        for block in self.SDDCNN:
            x = block(x)
            block_features.append(x)

        if self.use_mean_pooling:
            x = torch.mean(x, dim=[3, 4])
            x = x.permute(0, 2, 1)
        else:
            x = x.permute(0, 2, 3, 4, 1)
            x = x.reshape(x.shape[0], x.shape[1], -1)

        if self.frame_sim_layer is not None:
            x = torch.cat([self.frame_sim_layer(block_features), x], 2)

        if self.color_hist_layer is not None:
            x = torch.cat([self.color_hist_layer(inputs), x], 2)

        x = self.fc1(x)
        x = functional.relu(x)

        if self.dropout is not None:
            x = self.dropout(x)

        one_hot = self.cls_layer1(x)

        # scale from 0 to 1
        # one_hot
        return torch.sigmoid(one_hot)


class StackedDDCNNV2(nn.Module):
    """Stacked dilated dense convolutional neural network for video feature extraction."""

    def __init__(  # noqa: PLR0913
        self,
        in_filters: int,
        n_blocks: int,
        filters: int,
        *,
        shortcut: bool = True,
        pool_type: str = "avg",
        stochastic_depth_drop_prob: float = 0.0,
    ) -> None:
        """Initialize the stacked dilated dense convolutional network.

        Args:
            in_filters: Number of input filters.
            n_blocks: Number of blocks in the network.
            filters: Number of output filters.
            shortcut: Whether to use a shortcut connection.
            pool_type: Type of pooling to use.
            stochastic_depth_drop_prob: Dropout probability for stochastic depth.

        """
        super().__init__()
        if pool_type not in ("max", "avg"):
            msg = f"pool_type must be 'max' or 'avg', got {pool_type}"
            raise ValueError(msg)
        self.shortcut = shortcut
        self.DDCNN = nn.ModuleList(
            [
                DilatedDCNNV2(
                    in_filters if i == 1 else filters * 4,
                    filters,
                    activation=functional.relu if i != n_blocks else None,
                )
                for i in range(1, n_blocks + 1)
            ],
        )
        self.pool = nn.MaxPool3d(kernel_size=(1, 2, 2)) if pool_type == "max" else nn.AvgPool3d(kernel_size=(1, 2, 2))
        self.stochastic_depth_drop_prob = stochastic_depth_drop_prob

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Process input through the stacked dilated dense convolutional network.

        Args:
            inputs: Input tensor.

        Returns:
            Processed tensor.

        """
        x = inputs
        shortcut: torch.Tensor | None = None

        for block in self.DDCNN:
            x = block(x)
            if shortcut is None:
                shortcut = x
        if shortcut is None:
            msg = "shortcut should not be None"
            raise RuntimeError(msg)

        x = functional.relu(x)

        if self.shortcut is not None:
            if self.stochastic_depth_drop_prob != 0.0:
                if self.training:
                    x = shortcut if torch.rand(1).item() < self.stochastic_depth_drop_prob else x + shortcut
                else:
                    x = (1 - self.stochastic_depth_drop_prob) * x + shortcut
            else:
                x += shortcut

        return self.pool(x)  # type: ignore[no-any-return]


class DilatedDCNNV2(nn.Module):
    """Dilated dense convolutional model with multiple dilation rates."""

    def __init__(
        self,
        in_filters: int,
        filters: int,
        *,
        batch_norm: bool = True,
        activation: Callable[[torch.Tensor], torch.Tensor] | None = None,
    ) -> None:
        """Initialize the dilated dense convolutional model.

        Args:
            in_filters: Number of input filters.
            filters: Number of output filters.
            batch_norm: Whether to use batch normalization.
            activation: Activation function.

        """
        super().__init__()
        self.Conv3D_1 = Conv3DConfigurable(in_filters, filters, 1, use_bias=not batch_norm)
        self.Conv3D_2 = Conv3DConfigurable(in_filters, filters, 2, use_bias=not batch_norm)
        self.Conv3D_4 = Conv3DConfigurable(in_filters, filters, 4, use_bias=not batch_norm)
        self.Conv3D_8 = Conv3DConfigurable(in_filters, filters, 8, use_bias=not batch_norm)

        self.bn = nn.BatchNorm3d(filters * 4, eps=1e-3) if batch_norm else None
        self.activation = activation

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Process input through the dilated dense convolutional network.

        Args:
            inputs: Input tensor.

        Returns:
            Processed tensor.

        """
        conv1 = self.Conv3D_1(inputs)
        conv2 = self.Conv3D_2(inputs)
        conv3 = self.Conv3D_4(inputs)
        conv4 = self.Conv3D_8(inputs)

        x = torch.cat([conv1, conv2, conv3, conv4], dim=1)

        if self.bn is not None:
            x = self.bn(x)

        if self.activation is not None:
            x = self.activation(x)

        return x


class Conv3DConfigurable(nn.Module):
    """Configurable 3D convolution layer with support for separable convolutions."""

    def __init__(
        self,
        in_filters: int,
        filters: int,
        dilation_rate: int,
        *,
        separable: bool = True,
        use_bias: bool = True,
    ) -> None:
        """Initialize the 3D convolutional layer.

        Args:
            in_filters: Number of input filters.
            filters: Number of output filters.
            dilation_rate: Dilation rate for the convolution.
            separable: Whether to use separable convolution.
            use_bias: Whether to use bias in the convolution.

        """
        super().__init__()

        if separable:
            # (2+1)D convolution https://arxiv.org/pdf/1711.11248.pdf
            conv1 = nn.Conv3d(
                in_filters,
                2 * filters,
                kernel_size=(1, 3, 3),
                dilation=(1, 1, 1),
                padding=(0, 1, 1),
                bias=False,
            )
            conv2 = nn.Conv3d(
                2 * filters,
                filters,
                kernel_size=(3, 1, 1),
                dilation=(dilation_rate, 1, 1),
                padding=(dilation_rate, 0, 0),
                bias=use_bias,
            )
            self.layers = nn.ModuleList([conv1, conv2])
        else:
            conv = nn.Conv3d(
                in_filters,
                filters,
                kernel_size=3,
                dilation=(dilation_rate, 1, 1),
                padding=(dilation_rate, 1, 1),
                bias=use_bias,
            )
            self.layers = nn.ModuleList([conv])

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Process input through the 3D convolutional layers.

        Args:
            inputs: Input tensor.

        Returns:
            Processed tensor.

        """
        x = inputs
        for layer in self.layers:
            x = layer(x)
        return x


class FrameSimilarity(nn.Module):
    """Model for computing frame similarity features in video sequences."""

    def __init__(
        self,
        in_filters: int,
        similarity_dim: int = 128,
        lookup_window: int = 101,
        output_dim: int = 128,
        *,
        use_bias: bool = False,
    ) -> None:
        """Initialize the frame similarity model.

        Args:
            in_filters: Number of input filters.
            similarity_dim: Dimension of similarity features.
            lookup_window: Size of the window for similarity computation.
            output_dim: Dimension of the output features.
            use_bias: Whether to use bias in linear layers.

        """
        super().__init__()
        self.projection = nn.Linear(in_filters, similarity_dim, bias=use_bias)
        self.fc = nn.Linear(lookup_window, output_dim)

        self.lookup_window = lookup_window
        if lookup_window % 2 != 1:
            msg = "`lookup_window` must be odd integer"
            raise ValueError(msg)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Process input frames through the model.

        Args:
            inputs: Input tensor of video frames.

        Returns:
            Frame similarity features.

        """
        x = torch.cat([torch.mean(x, dim=[3, 4]) for x in inputs], dim=1)
        x = torch.transpose(x, 1, 2)

        x = self.projection(x)
        x = functional.normalize(x, p=2, dim=2)

        batch_size, time_window = x.shape[0], x.shape[1]
        similarities = torch.bmm(x, x.transpose(1, 2))  # [batch_size, time_window, time_window]
        similarities_padded = functional.pad(
            similarities,
            [(self.lookup_window - 1) // 2, (self.lookup_window - 1) // 2],
        )

        batch_indices = (
            torch.arange(0, batch_size, device=x.device)
            .view([batch_size, 1, 1])
            .repeat([1, time_window, self.lookup_window])
        )
        time_indices = (
            torch.arange(0, time_window, device=x.device)
            .view([1, time_window, 1])
            .repeat([batch_size, 1, self.lookup_window])
        )
        lookup_indices = (
            torch.arange(0, self.lookup_window, device=x.device)
            .view([1, 1, self.lookup_window])
            .repeat([batch_size, time_window, 1])
            + time_indices
        )

        similarities = similarities_padded[batch_indices, time_indices, lookup_indices]
        return functional.relu(self.fc(similarities))


class ColorHistograms(nn.Module):
    """Model for computing and comparing color histograms across video frames."""

    def __init__(self, lookup_window: int = 101, output_dim: int | None = None) -> None:
        """Initialize the color histogram model.

        Args:
            lookup_window: Size of the window for histogram computation.
            output_dim: Optional dimension for the output features.

        """
        super().__init__()

        self.fc = nn.Linear(lookup_window, output_dim) if output_dim is not None else None
        self.lookup_window = lookup_window
        if lookup_window % 2 != 1:
            msg = "`lookup_window` must be odd integer"
            raise ValueError(msg)

    @staticmethod
    def compute_color_histograms(frames: torch.Tensor) -> torch.Tensor:
        """Compute color histograms for video frames.

        Args:
            frames: Input tensor of video frames.

        Returns:
            Color histogram tensor.

        """
        frames = frames.int()

        num_chans: int = 3

        def get_bin(frames: torch.Tensor) -> torch.Tensor:
            """Get color bin indices for frames.

            Args:
                frames: Input tensor of video frames with RGB channels.

            Returns:
                Tensor of color bin indices, values are 0 .. 511

            """
            r, g, b = frames[:, :, 0], frames[:, :, 1], frames[:, :, 2]
            r, g, b = r >> 5, g >> 5, b >> 5
            return (r << 6) + (g << 3) + b

        batch_size, time_window, height, width, no_channels = frames.shape
        if no_channels != num_chans:
            msg = f"Expected {num_chans} channels, got {no_channels}"
            raise ValueError(msg)
        frames_flatten = frames.view(batch_size * time_window, height * width, 3)

        binned_values = get_bin(frames_flatten)
        frame_bin_prefix = (torch.arange(0, batch_size * time_window, device=frames.device) << 9).view(-1, 1)
        binned_values = (binned_values + frame_bin_prefix).view(-1)

        histograms = torch.zeros(batch_size * time_window * 512, dtype=torch.int32, device=frames.device)
        histograms.scatter_add_(
            0,
            binned_values,
            torch.ones(len(binned_values), dtype=torch.int32, device=frames.device),
        )

        histograms = histograms.view(batch_size, time_window, 512).float()
        # histograms_normalized
        return functional.normalize(histograms, p=2, dim=2)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Process input frames through the model.

        Args:
            inputs: Input tensor of video frames.

        Returns:
            Model predictions for shot transitions.

        """
        x = self.compute_color_histograms(inputs)

        batch_size, time_window = x.shape[0], x.shape[1]
        similarities = torch.bmm(x, x.transpose(1, 2))  # [batch_size, time_window, time_window]
        similarities_padded = functional.pad(
            similarities,
            [(self.lookup_window - 1) // 2, (self.lookup_window - 1) // 2],
        )

        batch_indices = (
            torch.arange(0, batch_size, device=x.device)
            .view([batch_size, 1, 1])
            .repeat([1, time_window, self.lookup_window])
        )
        time_indices = (
            torch.arange(0, time_window, device=x.device)
            .view([1, time_window, 1])
            .repeat([batch_size, 1, self.lookup_window])
        )
        lookup_indices = (
            torch.arange(0, self.lookup_window, device=x.device)
            .view([1, 1, self.lookup_window])
            .repeat([batch_size, time_window, 1])
            + time_indices
        )

        similarities = similarities_padded[batch_indices, time_indices, lookup_indices]

        if self.fc is not None:
            return functional.relu(self.fc(similarities))
        return similarities


class TransNetV2(ModelInterface):
    """Interface for TransNetV2 shot transition detection model."""

    def __init__(self, model_dir: str | None = None) -> None:
        """Initialize the TransNetV2 model interface."""
        super().__init__()
        self.model_dir = model_dir

    @property
    def model_id_names(self) -> list[str]:
        """Get the model ID names.

        Returns:
            A list of model ID names.

        """
        return [_TRANSNETV2_MODEL_ID]

    def setup(self) -> None:
        """Set up the TransNetV2 model interface."""
        self._model = _TransNetV2()
        model_dir = pathlib.Path(self.model_dir) / _TRANSNETV2_MODEL_ID
        model_file = model_dir / _TRANSNETV2_MODEL_WEIGHTS
        if not model_file.exists():
            error_msg = f"{model_file} not found!"
            raise FileNotFoundError(error_msg)
        state_dict = torch.load(model_file.as_posix())
        self._model.load_state_dict(state_dict)
        self._model.eval().cuda()

    def __call__(self, inputs: torch.Tensor) -> torch.Tensor:
        """TransNetV2 model call.

        Args:
            inputs: tensor of shape [# batch, # frames, height, width, RGB].

        Returns:
            tensor of shape [# batch, # frames, 1] of probabilities for each frame being a shot transition.

        """
        with torch.no_grad():
            return self._model(inputs)  # type: ignore[no-any-return]

    @classmethod
    def download_weights_on_node(cls, model_dir: str) -> None:
        """Download TransNetV2 weights on the node.

        Args:
            model_dir: Directory to save the model weights. If None, uses self.model_dir.
        """

        # Create the model directory if it doesn't exist
        model_dir_path = pathlib.Path(model_dir) / _TRANSNETV2_MODEL_ID
        model_dir_path.mkdir(parents=True, exist_ok=True)
        model_file = model_dir_path / _TRANSNETV2_MODEL_WEIGHTS
        if model_file.exists():
            return

        # Download the weights file from Hugging Face
        download_model_from_hf(
            model_id=_TRANSNETV2_MODEL_ID,
            local_dir=model_dir_path,
            filename=_TRANSNETV2_MODEL_WEIGHTS,
            revision=_TRANSNETV2_MODEL_REVISION,
        )

        logger.info(f"TransNetV2 weights downloaded to: {model_file}")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

import editdistance

from nemo_curator.stages.audio.common import LegacySpeechStage
from nemo_curator.tasks import AudioBatch


def get_wer(text: str, pred_text: str) -> float:
    text_words = text.split()
    pred_text_words = pred_text.split()
    word_dist = editdistance.eval(text_words, pred_text_words)

    num_words = len(text_words)
    return round(word_dist / num_words * 100.0, 2)


def get_cer(text: str, pred_text: str) -> float:
    char_dist = editdistance.eval(text, pred_text)
    num_chars = len(text)
    return round(char_dist / num_chars * 100.0, 2)


def get_charrate(text: str, duration: float) -> float:
    num_chars = len(text)
    return round(num_chars / duration, 2)


def get_wordrate(text: str, duration: float) -> float:
    num_words = len(text.split())
    return round(num_words / duration, 2)


@dataclass
class GetPairwiseWerStage(LegacySpeechStage):
    """Count pairwise word-error-rate (WER) * 100% for each pair of text and pred_text.

    WER is measured between ``data[self.text_key]`` and ``data[self.pred_text_key]``.


    Args:
        text_key (str): a string indicating which key of the data entries
            should be used to find the utterance transcript. Defaults to "text".
        pred_text_key (str): a string indicating which key of the data entries
            should be used to access the ASR predictions. Defaults to "pred_text".

    Returns:
         The same data as in the input manifest with wer_key and corresponding values.
    """

    text_key: str = "text"
    pred_text_key: str = "pred_text"
    wer_key: str = "wer"

    def process_dataset_entry(self, data_entry: dict) -> list[AudioBatch]:
        wer = get_wer(data_entry[self.text_key], data_entry[self.pred_text_key])
        data_entry[self.wer_key] = wer
        return [AudioBatch(data=data_entry)]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

import numpy as np
import torch
from loguru import logger

from nemo_curator.backends.base import NodeInfo, WorkerMetadata
from nemo_curator.models.nsfw import NSFWScorer
from nemo_curator.stages.image.filters.base import BaseFilterStage
from nemo_curator.tasks import ImageBatch


@dataclass
class ImageNSFWFilterStage(BaseFilterStage):
    """Stage for filtering out NSFW images using NSFWScorer model.

    This class processes image batches through an NSFW scoring model to generate
    NSFW probability scores for each image. Images with scores above the threshold
    will be filtered out as NSFW content.
    """
    weights_path: str = None
    _name: str = "image_nsfw_filter"

    def setup_on_node(self, _node_info: NodeInfo | None = None, _worker_metadata: WorkerMetadata | None = None) -> None:
        """Download NSFW model weights from LAION repository."""
        NSFWScorer.download_weights_on_node(self.model_dir)

    def setup(self, _worker_metadata: WorkerMetadata | None = None) -> None:
        """Initialize the NSFW filtering model."""
        self.model = NSFWScorer(model_dir=self.model_dir)
        self.model.setup()

        if self.verbose:
            logger.info("Initialized NSFW scoring model")

    def process(self, task: ImageBatch) -> ImageBatch:
        """Process an image batch to generate NSFW scores and filter by threshold.

        Args:
            task: ImageBatch containing list of ImageObject instances with pre-computed embeddings

        Returns:
            ImageBatch with filtered images that have NSFW scores below the threshold
        """

        # Process images in batches to generate scores
        for batch in self.yield_next_batch(task):
            # Stack embeddings into batch tensor (N, embedding_dim)
            embeddings = [img_obj.embedding for img_obj in batch]
            batch_tensor = np.stack(embeddings, axis=0)

            # Generate NSFW scores
            with torch.no_grad():
                scores = self.model(batch_tensor).cpu().numpy()

            # Store scores in ImageObject.nsfw_score
            for i, image_obj in enumerate(batch):
                image_obj.nsfw_score = float(scores[i])

            if self.verbose:
                logger.info(
                    f"Generated NSFW scores for {len(batch)} images "
                    f"in batch {i}-{i + self.model_inference_batch_size}"
                )

        # Filter images based on NSFW score threshold
        filtered_images = []
        filtered_count = 0

        for image_obj in task.data:
            if image_obj.nsfw_score < self.score_threshold:
                filtered_images.append(image_obj)
            else:
                filtered_count += 1
                if self.verbose:
                    logger.info(
                        f"Image {image_obj.image_id} (path: {image_obj.image_path}) has NSFW score {image_obj.nsfw_score:.3f} "
                        f"above threshold {self.score_threshold}, filtered out as NSFW."
                    )

        if self.verbose:
            logger.info(
                f"NSFW filtering: {len(filtered_images)}/{len(task.data)} images passed, "
                f"{filtered_count} filtered out"
            )

        # Return new ImageBatch with filtered images
        return ImageBatch(
            data=filtered_images,
            dataset_name=task.dataset_name,
            task_id=f"{task.task_id}_{self.name}",
            _metadata=task._metadata,
            _stage_perf=task._stage_perf,
        )


# Explicitly export the class
__all__ = ["ImageNSFWFilterStage"]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from nemo_curator.stages.base import ProcessingStage
from nemo_curator.tasks import AudioBatch, DocumentBatch


class AudioToDocumentStage(ProcessingStage[AudioBatch, DocumentBatch]):
    """
    Stage to conver DocumentObject to DocumentBatch

    """

    def process(self, task: AudioBatch) -> list[DocumentBatch]:
        return [
            DocumentBatch(
                data=pd.DataFrame(task.data),
                task_id=task.task_id,
                dataset_name=task.dataset_name,
            )
        ]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import TYPE_CHECKING, Any

import cudf
from cugraph.dask.comms.comms_wrapper import init_subcomms as c_init_subcomms
from loguru import logger
from pylibcugraph import GraphProperties, MGGraph, ResourceHandle
from pylibcugraph import weakly_connected_components as pylibcugraph_wcc

from nemo_curator.backends.experimental.utils import RayStageSpecKeys
from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.deduplication.fuzzy.utils import CURATOR_FUZZY_DUPLICATE_GROUP_FIELD
from nemo_curator.stages.deduplication.id_generator import (
    CURATOR_DEDUP_ID_STR,
)
from nemo_curator.stages.deduplication.io_utils import DeduplicationIO
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks.file_group import FileGroupTask
from nemo_curator.utils.file_utils import create_or_overwrite_dir, get_fs

if TYPE_CHECKING:
    from nemo_curator.backends.base import WorkerMetadata


class ConnectedComponentsStage(ProcessingStage[FileGroupTask, FileGroupTask], DeduplicationIO):
    def __init__(
        self,
        output_path: str,
        source_field: str = f"{CURATOR_DEDUP_ID_STR}_x",
        destination_field: str = f"{CURATOR_DEDUP_ID_STR}_y",
        read_kwargs: dict | None = None,
        write_kwargs: dict | None = None,
    ):
        """
        Args:
            output_path: The path to write the resulting connected components to.
            source_field: The field name containing the document ids of the source of the edge.
            destination_field: The field name containing the document ids of the destination of the edge.
            read_kwargs: Keyword arguments to pass for reading the input files.
            write_kwargs: Keyword arguments to pass for writing the output files.
        """

        self.source_field = source_field
        self.destination_field = destination_field
        self.read_kwargs = read_kwargs if read_kwargs is not None else {}
        self.write_kwargs = write_kwargs if write_kwargs is not None else {}

        self._name = self.__class__.__name__
        self._resources = Resources(cpus=1.0, gpus=1.0)
        self._batch_size = None

        # Handle output directory cleanup logic
        self.output_fs = get_fs(output_path, self.write_kwargs.get("storage_options"))
        self.output_path = self.output_fs.sep.join([output_path, self.name])
        create_or_overwrite_dir(self.output_path, fs=self.output_fs)

    def setup(self, _worker_metadata: "WorkerMetadata | None" = None) -> None:
        if not hasattr(self, "_raft_handle"):
            msg = "RAFT handle not found. Make sure the stage is initialized with RAFT"
            raise ValueError(msg)

        self._setup_post()

    def ray_stage_spec(self) -> dict[str, Any]:
        return {
            RayStageSpecKeys.IS_RAFT_ACTOR: True,
        }

    def __get_2D_div(self, ngpus: int) -> tuple[int, int]:  # noqa: N802
        """
        Cugraph 2d partitioning number of rows and columns
        """
        # Taken from https://github.com/rapidsai/cugraph/blob/branch-25.06/python/cugraph/cugraph/dask/comms/comms.py#L41-L45
        prows = int(math.sqrt(ngpus))
        while ngpus % prows != 0:
            prows = prows - 1
        return prows, int(ngpus / prows)

    def _setup_post(self) -> None:
        """Setup the sub-communicator for cuGraph communications.

        This method is specific to cuGraph comms and is used to initialize the
        sub-communicator.
        """
        if not hasattr(self, "_raft_handle") or not hasattr(self, "_actor_pool_size"):
            msg = "RAFT handle or actor pool size not found. Make sure the stage is initialized with RAFT"
            raise ValueError(msg)

        logger.debug("     Setting up cuGraph-subcom...")
        row_comm_size, _ = self.__get_2D_div(self._actor_pool_size)
        c_init_subcomms(self._raft_handle, row_comm_size)

    def weakly_connected_components(self, df: cudf.DataFrame, src_col: str, dst_col: str) -> None:
        """Compute the weakly connected components of a graph.

        This method loads a chunk of the graph, creates a cuGraph object, and
        computes the weakly connected components using the MGGraph library.

        Parameters
        ----------
        start: int
            The start index of the chunk.
        stop: int
            The stop index of the chunk.
        """

        src_array = df[src_col]
        dst_array = df[dst_col]

        rhandle = ResourceHandle(self._raft_handle.getHandle())

        graph_props = GraphProperties(
            is_multigraph=False,
            is_symmetric=True,
        )
        logger.debug("Running graph creation")
        plc_graph = MGGraph(
            resource_handle=rhandle,
            graph_properties=graph_props,
            src_array=[src_array],
            dst_array=[dst_array],
            edge_id_array=None,
            edge_type_array=None,
            num_arrays=1,
            store_transposed=False,
            symmetrize=False,
            do_expensive_check=False,
            drop_multi_edges=True,
        )
        logger.debug("Running weakly connected components")
        res = pylibcugraph_wcc(
            resource_handle=rhandle,
            graph=plc_graph,
            offsets=None,
            indices=None,
            weights=None,
            labels=None,
            do_expensive_check=False,
        )
        logger.info("Computing weakly connected components completed successfully!")
        return res

    def process(self, task: FileGroupTask) -> FileGroupTask:
        err_msg = "ConnectedComponentsStage only support process batch"
        raise NotImplementedError(err_msg)

    def process_batch(self, tasks: list[FileGroupTask]) -> list[FileGroupTask]:
        """
        Process a batch of input files containing edges between documents.
        Compute the weakly connected components of the graph and write a mapping of document ids to their connected component id.

        Parameters
        ----------
        tasks: list[FileGroupTask]
            A list of FileGroupTasks containing the input files.
        Returns
        -------
        list[FileGroupTask]
            A list of FileGroupTasks containing the output doc_id to connected component id mapping.
        """
        input_files = []
        for task in tasks:
            input_files.extend(task.data)
        output_file = self.output_fs.sep.join([self.output_path, f"{tasks[0]._uuid}.parquet"])
        edgelist_columns = [self.source_field, self.destination_field]
        dfs = []
        for input_file in input_files:
            dfs.append(self.read_parquet(input_file, columns=edgelist_columns, **self.read_kwargs))
        df = cudf.concat(dfs)
        # remove duplicate edges
        df = df.drop_duplicates(subset=edgelist_columns, ignore_index=True)
        vertices, labels = self.weakly_connected_components(df, self.source_field, self.destination_field)
        df = cudf.DataFrame(
            {
                CURATOR_DEDUP_ID_STR: vertices,
                CURATOR_FUZZY_DUPLICATE_GROUP_FIELD: labels,
            }
        )
        self.write_parquet(df=df, filepath=output_file, index=False, **self.write_kwargs)
        return [
            FileGroupTask(
                dataset_name=tasks[0].dataset_name,
                task_id=tasks[0].task_id,
                data=[output_file],
                _metadata={
                    "storage_options": self.write_kwargs.get("storage_options"),
                    "num_vertices": len(vertices),
                },
            )
        ]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import os
from collections.abc import Generator
from typing import Literal

os.environ["RAPIDS_NO_INITIALIZE"] = "1"

import numpy as np
import pandas as pd
import torch
from huggingface_hub import snapshot_download
from loguru import logger

from nemo_curator.backends.base import NodeInfo, WorkerMetadata
from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks import DocumentBatch

from .utils import ATTENTION_MASK_COLUMN, INPUT_ID_COLUMN, SEQ_ORDER_COLUMN, clip_tokens, format_name_with_suffix


class ModelStage(ProcessingStage[DocumentBatch, DocumentBatch]):
    """
    Base class for Hugging Face model inference.

    Args:
        model_identifier: The identifier of the Hugging Face model.
        cache_dir: The Hugging Face cache directory. Defaults to None.
        hf_token: Hugging Face token for downloading the model, if needed. Defaults to None.
        model_inference_batch_size: The size of the batch for model inference. Defaults to 256.
        has_seq_order: Whether to sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        padding_side: The side to pad the input tokens. Defaults to "right".
        unpack_inference_batch: Whether to unpack the inference batch with **kwargs. Defaults to False.
        autocast: Whether to use autocast. When True, we trade off minor accuracy for faster inference.
            Defaults to True.

    """

    def __init__(  # noqa: PLR0913
        self,
        model_identifier: str,
        cache_dir: str | None = None,
        hf_token: str | None = None,
        model_inference_batch_size: int = 256,
        has_seq_order: bool = True,
        padding_side: Literal["left", "right"] = "right",
        unpack_inference_batch: bool = False,
        autocast: bool = True,
    ):
        self._name = format_name_with_suffix(model_identifier, suffix="_model")
        # Assume that the model can fit on a single GPU
        self._resources = Resources(cpus=1, gpus=1)

        self.model_identifier = model_identifier
        self.cache_dir = cache_dir
        self.hf_token = hf_token
        self.model_inference_batch_size = model_inference_batch_size
        self.has_seq_order = has_seq_order
        self.padding_side = padding_side
        self.unpack_inference_batch = unpack_inference_batch
        self.autocast = autocast

    def inputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [INPUT_ID_COLUMN, ATTENTION_MASK_COLUMN] + ([SEQ_ORDER_COLUMN] if self.has_seq_order else [])

    def outputs(self) -> tuple[list[str], list[str]]:
        msg = "Subclasses must implement this method"
        raise NotImplementedError(msg)

    def setup_on_node(self, _node_info: NodeInfo | None = None, _worker_metadata: WorkerMetadata = None) -> None:
        try:
            snapshot_download(
                repo_id=self.model_identifier,
                cache_dir=self.cache_dir,
                token=self.hf_token,
                local_files_only=False,
            )

            _setup_function = getattr(self, "_setup", None)
            if callable(_setup_function):
                _setup_function(local_files_only=False)
            else:
                logger.warning(f"Subclass {self.__class__.__name__} does not implement _setup method")

        except Exception as e:
            msg = f"Failed to download {self.model_identifier}"
            raise RuntimeError(msg) from e

    def setup(self, _: WorkerMetadata | None = None) -> None:
        _setup_function = getattr(self, "_setup", None)
        if callable(_setup_function):
            # We use the _setup function to ensure that everything needed for the model is downloaded and loaded properly
            _setup_function(local_files_only=True)
        else:
            msg = "Subclasses must implement this method"
            raise NotImplementedError(msg)

    def yield_next_batch(self, df: pd.DataFrame) -> Generator[dict[str, torch.Tensor]]:
        """
        Yields a generator of model inputs for the next batch.
        We only move the batch to the GPU to reduce the memory overhead.

        Args:
            df (pd.DataFrame): The Pandas DataFrame (with input_ids and attention_mask) to process.

        Yields:
            Generator[dict[str, torch.Tensor]]: A generator of model inputs for the next batch.

        """
        for i in range(0, len(df), self.model_inference_batch_size):
            yield clip_tokens(
                {
                    INPUT_ID_COLUMN: torch.tensor(
                        df[INPUT_ID_COLUMN][i : i + self.model_inference_batch_size].tolist()
                    ).to(self.model.device),
                    ATTENTION_MASK_COLUMN: torch.tensor(
                        df[ATTENTION_MASK_COLUMN][i : i + self.model_inference_batch_size].tolist()
                    ).to(self.model.device),
                },
                padding_side=self.padding_side,
            )

    def process_model_output(
        self, outputs: torch.Tensor, model_input_batch: dict[str, torch.Tensor] | None = None
    ) -> dict[str, np.ndarray] | torch.Tensor:
        msg = "Subclasses must implement this method"
        raise NotImplementedError(msg)

    def collect_outputs(self, processed_outputs: list[dict[str, np.ndarray]]) -> dict[str, np.ndarray]:
        result = {}
        for key in processed_outputs[0]:
            result[key] = np.concatenate([out[key] for out in processed_outputs], axis=0)
        return result

    def create_output_dataframe(self, df_cpu: pd.DataFrame, collected_output: dict[str, np.ndarray]) -> pd.DataFrame:
        msg = "Subclasses must implement this method"
        raise NotImplementedError(msg)

    def _model_forward(self, model_input_batch: dict[str, torch.Tensor]) -> torch.Tensor:
        if self.unpack_inference_batch:
            return self.model(**model_input_batch)
        else:
            return self.model(model_input_batch)

    def process(self, batch: DocumentBatch) -> DocumentBatch:
        df_cpu = batch.to_pandas()

        processed_outputs = []
        for model_input_batch in self.yield_next_batch(df_cpu):
            # Forward pass
            with torch.no_grad():
                if self.autocast:
                    with torch.autocast(device_type="cuda"):
                        outputs = self._model_forward(model_input_batch)
                else:
                    outputs = self._model_forward(model_input_batch)

            processed_output = self.process_model_output(outputs, model_input_batch)
            del model_input_batch
            processed_outputs.append(processed_output)

        # Collect all outputs
        collected_output = self.collect_outputs(processed_outputs)

        # Create output Pandas DataFrame
        df_cpu = self.create_output_dataframe(df_cpu, collected_output)

        # Sort by seq_order to preserve original order from tokenizer
        if self.has_seq_order:
            df_cpu = df_cpu.sort_values(by=SEQ_ORDER_COLUMN, ignore_index=True).drop(columns=[SEQ_ORDER_COLUMN])

        return DocumentBatch(
            task_id=batch.task_id,
            dataset_name=batch.dataset_name,
            data=df_cpu,
            _metadata=batch._metadata,
            _stage_perf=batch._stage_perf,
        )

    def teardown(self) -> None:
        gc.collect()
        torch.cuda.empty_cache()



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from dataclasses import dataclass
from operator import eq, ge, gt, le, lt, ne

import soundfile
from loguru import logger

from nemo_curator.stages.base import ProcessingStage
from nemo_curator.tasks import AudioBatch, Task


class LegacySpeechStage(ProcessingStage[Task, Task]):
    """
    LegacySpeechStage for SDP processors inherited from BaseParallelProcessor

    """

    def process(self, task: AudioBatch) -> list[Task]:
        result = []
        for entry in task.data:
            result.extend(self.process_dataset_entry(entry))
        return result

    @abstractmethod
    def process_dataset_entry(self, data_entry: AudioBatch) -> list[AudioBatch]:
        return [data_entry]


@dataclass
class GetAudioDurationStage(LegacySpeechStage):
    """
    Stage that computes the duration of the file in ``audio_filepath_key`` (using soundfile)
    and saves the duration in ``duration_key``. If there is an error computing the duration,
    the value at ``duration_key`` will be updated with the value -1.0.

    Args:
        audio_filepath_key (str): Key to get path to wav file.
        duration_key (str): Key to put to audio duration.
    Returns:
        All the same fields as in the input manifest plus duration_key
    """

    audio_filepath_key: str
    duration_key: str

    def process_dataset_entry(self, data_entry: dict) -> list[AudioBatch]:
        audio_filepath = data_entry[self.audio_filepath_key]
        try:
            data, samplerate = soundfile.read(audio_filepath)
            data_entry[self.duration_key] = data.shape[0] / samplerate
        except soundfile.SoundFileError as e:
            logger.warning(str(e) + " file: " + audio_filepath)
            data_entry[self.duration_key] = -1.0
        return [AudioBatch(data=data_entry)]


class PreserveByValueStage(LegacySpeechStage):
    """
    Processor for preserving dataset entries based on a specified condition involving a target value and an input field.

    Args:
        input_value_key (str): The field in the dataset entries to be evaluated.
        target_value (Union[int, str]): The value to compare with the input field.
        operator (str): (Optional) The operator to apply for comparison. Options: "lt" (less than), "le" (less than or equal to), "eq" (equal to), "ne" (not equal to), "ge" (greater than or equal to), "gt" (greater than). Defaults to "eq".
        **kwargs: Additional keyword arguments to be passed to the base class `BaseParallelProcessor`.

    """

    def __init__(
        self,
        input_value_key: str,
        target_value: int | str,
        operator: str = "eq",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.input_value_key = input_value_key
        self.target_value = target_value
        if operator == "lt":
            self.operator = lt
        elif operator == "le":
            self.operator = le
        elif operator == "eq":
            self.operator = eq
        elif operator == "ne":
            self.operator = ne
        elif operator == "ge":
            self.operator = ge
        elif operator == "gt":
            self.operator = gt
        else:
            msg = 'Operator must be one from the list: "lt" (less than), "le" (less than or equal to), "eq" (equal to), "ne" (not equal to), "ge" (greater than or equal to), "gt" (greater than)'
            raise ValueError(msg)

    def process_dataset_entry(self, data_entry: AudioBatch) -> list[AudioBatch]:
        input_value = data_entry[self.input_value_key]
        target = self.target_value
        if self.operator(input_value, target):
            return [AudioBatch(data=data_entry)]
        else:
            return []



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from nemo_curator.backends.experimental.utils import get_available_cpu_gpu_resources
from nemo_curator.stages.base import ProcessingStage


def calculate_concurrency_for_actors_for_stage(stage: ProcessingStage) -> tuple[int, int] | int:
    """
    Calculate concurrency if we want to spin up actors based on available resources and stage requirements.

    Returns:
        int | tuple[int, int]: Number of actors to use
            int: Number of workers to use
            tuple[int, int]: tuple of min / max actors to use and number of workers to use
    """
    # If explicitly set, use the specified number of workers
    num_workers = stage.num_workers()
    if num_workers is not None and num_workers > 0:
        return max(1, num_workers)

    # Get available resources from Ray
    available_cpus, available_gpus = get_available_cpu_gpu_resources(init_and_shudown=False)
    # Calculate based on CPU and GPU requirements
    max_cpu_actors = float("inf")
    max_gpu_actors = float("inf")

    # CPU constraint
    if stage.resources.cpus > 0:
        max_cpu_actors = available_cpus // stage.resources.cpus

    # GPU constraint
    if stage.resources.gpus > 0:
        max_gpu_actors = available_gpus // stage.resources.gpus

    # Take the minimum of CPU and GPU constraints
    max_actors = min(max_cpu_actors, max_gpu_actors)
    return (1, int(max_actors))


def is_actor_stage(stage: ProcessingStage) -> bool:
    """Check if the stage is an actor stage."""
    overridden_setup = type(stage).setup is not ProcessingStage.setup
    has_gpu_and_cpu = (stage.resources.gpus > 0) and (stage.resources.cpus > 0)
    return overridden_setup or has_gpu_and_cpu



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Generator
from dataclasses import dataclass

import torch

from nemo_curator.backends.base import WorkerMetadata
from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.resources import Resources
from nemo_curator.tasks import ImageBatch, ImageObject


@dataclass
class BaseFilterStage(ProcessingStage[ImageBatch, ImageBatch]):
    """Base class for image filtering stages.

    This class provides a base class for image filtering stages.
    """
    model_dir: str = None
    num_gpus_per_worker: float = 0.25
    model_inference_batch_size: int = 32  # Number of images to process through model at once
    score_threshold: float = 0.5
    verbose: bool = False
    _name: str = "image_filter"

    def __post_init__(self) -> None:
        if torch.cuda.is_available():
            self._resources = Resources(gpus=self.num_gpus_per_worker)
        else:
            self._resources = Resources()

    def inputs(self) -> tuple[list[str], list[str]]:
        return ["data"], []

    def outputs(self) -> tuple[list[str], list[str]]:
        return ["data"], []

    def setup(self, _worker_metadata: WorkerMetadata | None = None) -> None:
        """Initialize the base filter stage."""
        raise NotImplementedError

    def yield_next_batch(self, task: ImageBatch) -> Generator[list[ImageObject], None, None]:
        """
        Yields a generator of model inputs for the next batch.

        Args:
            task (ImageBatch): The ImageBatch to process.

        Yields:
            Generator[dict[str, torch.Tensor]]: A generator of model inputs for the next batch.

        """
        for i in range(0, len(task.data), self.model_inference_batch_size):
            yield task.data[i : i + self.model_inference_batch_size]

    def process(self, task: ImageBatch) -> ImageBatch:
        """Process an image batch to generate scores and filter by threshold.

        Args:
            task: ImageBatch containing list of ImageObject instances with pre-computed embeddings

        Returns:
            ImageBatch with filtered images that have scores below the threshold
        """
        raise NotImplementedError


# Explicitly export the class
__all__ = ["BaseFilterStage"]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Literal

from nemo_curator.stages.base import ProcessingStage
from nemo_curator.stages.text.download import DocumentDownloadExtractStage
from nemo_curator.stages.text.download.html_extractors import HTMLExtractorAlgorithm

from .download import CommonCrawlWARCDownloader
from .extract import CommonCrawlHTMLExtractor
from .url_generation import MainCommonCrawlUrlGenerator, NewsCommonCrawlUrlGenerator
from .warc_iterator import CommonCrawlWarcIterator


class CommonCrawlDownloadExtractStage(DocumentDownloadExtractStage):
    """Composite stage for downloading and processing Common Crawl data.

    This pipeline:
    1. Generates WARC URLs (either from main or news crawls)
    2. Downloads WARC files
    3. Extracts content from WARC files
    4. Extracts text from HTML content
    """

    def __init__(  # noqa: PLR0913
        self,
        start_snapshot: str,
        end_snapshot: str,
        download_dir: str,
        crawl_type: Literal["main", "news"] = "main",
        html_extraction: HTMLExtractorAlgorithm | str | None = None,
        html_extraction_kwargs: dict | None = None,
        stop_lists: dict[str, frozenset[str]] | None = None,
        use_aws_to_download: bool = False,
        verbose: bool = False,
        url_limit: int | None = None,
        record_limit: int | None = None,
        add_filename_column: bool | str = True,
    ):
        self.crawl_type = crawl_type
        self.start_snapshot = start_snapshot
        self.end_snapshot = end_snapshot

        if crawl_type == "main":
            self.url_generator = MainCommonCrawlUrlGenerator(
                start_snapshot_str=start_snapshot, end_snapshot_str=end_snapshot, limit=url_limit
            )
        else:
            self.url_generator = NewsCommonCrawlUrlGenerator(
                start_snapshot_str=start_snapshot, end_snapshot_str=end_snapshot, limit=url_limit
            )

        self.downloader = CommonCrawlWARCDownloader(
            download_dir=download_dir, use_aws_to_download=use_aws_to_download, verbose=verbose
        )
        self.iterator = CommonCrawlWarcIterator()
        self.extractor = CommonCrawlHTMLExtractor(
            algorithm=html_extraction,
            algorithm_kwargs=html_extraction_kwargs,
            stop_lists=stop_lists,
        )
        super().__init__(
            url_generator=self.url_generator,
            downloader=self.downloader,
            iterator=self.iterator,
            extractor=self.extractor,
            url_limit=url_limit,
            record_limit=record_limit,
            add_filename_column=add_filename_column,
        )
        self._name = f"common_crawl_{self.crawl_type}_pipeline"

    def decompose(self) -> list[ProcessingStage]:
        """Decompose this composite stage into its constituent stages."""
        return self.stages

    def get_description(self) -> str:
        """Get a description of this composite stage."""
        return f"Common Crawl {self.crawl_type} pipeline: {self.start_snapshot} to {self.end_snapshot}"



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

os.environ["RAPIDS_NO_INITIALIZE"] = "1"
from dataclasses import dataclass
from typing import Any, Literal

import numpy as np
import pandas as pd
import torch
import torch.nn.functional as F  # noqa: N812
from huggingface_hub import PyTorchModelHubMixin, snapshot_download
from torch import nn
from torch.nn import Dropout, Linear
from transformers import AutoModelForCausalLM, AutoTokenizer

from nemo_curator.backends.base import NodeInfo, WorkerMetadata
from nemo_curator.stages.base import CompositeStage, ProcessingStage
from nemo_curator.stages.text.models.model import ModelStage
from nemo_curator.stages.text.models.tokenizer import TokenizerStage
from nemo_curator.stages.text.models.utils import ATTENTION_MASK_COLUMN, INPUT_ID_COLUMN, format_name_with_suffix
from nemo_curator.stages.text.modules.score_filter import Filter
from nemo_curator.tasks import DocumentBatch

from .aegis_utils import AEGIS_LABELS, format_aegis

PRETRAINED_MODEL_NAME_OR_PATH = "meta-llama/LlamaGuard-7b"
AEGIS_VARIANTS = [
    "nvidia/Aegis-AI-Content-Safety-LlamaGuard-Defensive-1.0",
    "nvidia/Aegis-AI-Content-Safety-LlamaGuard-Permissive-1.0",
]
INSTRUCTION_DATA_GUARD_MODEL_IDENTIFIER = "nvidia/instruction-data-guard"
HIDDEN_TEXT_COLUMN = "_curator_hidden_text"
MAX_SEQ_LENGTH = 4096
TOKENIZER_PADDING_SIDE = "left"
TORCH_DTYPE = torch.bfloat16


class InstructionDataGuardNet(torch.nn.Module, PyTorchModelHubMixin):
    def __init__(self, input_dim: int, dropout: float = 0.7):
        super().__init__()
        self.input_dim = input_dim
        self.dropout = Dropout(dropout)
        self.sigmoid = torch.nn.Sigmoid()
        self.input_layer = Linear(input_dim, input_dim)

        self.hidden_layer_0 = Linear(input_dim, 2000)
        self.hidden_layer_1 = Linear(2000, 500)
        self.hidden_layer_2 = Linear(500, 1)

    @torch.no_grad()
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.normalize(x, dim=-1)
        x = self.dropout(x)
        x = F.relu(self.input_layer(x))
        x = self.dropout(x)
        x = F.relu(self.hidden_layer_0(x))
        x = self.dropout(x)
        x = F.relu(self.hidden_layer_1(x))
        x = self.dropout(x)
        x = self.hidden_layer_2(x)
        return self.sigmoid(x)


class AegisModel(nn.Module):
    def __init__(  # noqa: PLR0913
        self,
        pretrained_model_name_or_path: str,
        peft_model_name_or_path: str,
        dtype: torch.dtype = TORCH_DTYPE,
        cache_dir: str | None = None,
        local_files_only: bool = True,
        hf_token: str | bool | None = None,
        add_instruction_data_guard: bool = False,
    ):
        super().__init__()

        base_model = AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path,
            torch_dtype=dtype,
            cache_dir=cache_dir,
            local_files_only=local_files_only,
            token=hf_token,
        )
        # Importing PeftModel here to prevent cuda context issues
        # that seem to happen on Transformers 4.48.3
        # See related: https://github.com/rapidsai/crossfit/pull/113
        from peft import PeftModel

        self.model = PeftModel.from_pretrained(
            base_model,
            peft_model_name_or_path,
            cache_dir=cache_dir,
            local_files_only=local_files_only,
        )
        self.add_instruction_data_guard = add_instruction_data_guard
        if self.add_instruction_data_guard:
            self.instruction_data_guard_net = InstructionDataGuardNet(4096)

    @property
    def device(self) -> torch.device:
        return next(self.parameters()).device

    @torch.no_grad()
    def forward(self, batch: dict[str, torch.Tensor]) -> torch.Tensor:
        batch = {k: v.to(TORCH_DTYPE) if v.dtype.is_floating_point else v for k, v in batch.items()}

        if self.add_instruction_data_guard:
            response = self.model.generate(
                **batch,
                max_new_tokens=1,
                pad_token_id=0,
                output_hidden_states=True,
                return_dict_in_generate=True,
            )
            # Access the hidden state of the last non-generated token from the last layer
            instruction_data_guard_input_tensor = response.hidden_states[0][32][:, -1, :].to(torch.float)

            del batch, response

            return self.instruction_data_guard_net(instruction_data_guard_input_tensor).flatten()
        else:
            response = self.model.generate(
                **batch,
                max_new_tokens=100,
                pad_token_id=0,
            )

            del batch

            return response


class AegisModelStage(ModelStage):
    """
    See ModelStage for more information.
    """

    def __init__(  # noqa: PLR0913
        self,
        model_identifier: str,
        cache_dir: str | None = None,
        hf_token: str | None = None,
        pred_column: str = "preds",
        prob_column: str = "probs",
        model_inference_batch_size: int = 256,
        has_seq_order: bool = True,
        add_instruction_data_guard: bool = False,
        autocast: bool = True,
    ):
        super().__init__(
            model_identifier=model_identifier,
            cache_dir=cache_dir,
            hf_token=hf_token,
            model_inference_batch_size=model_inference_batch_size,
            has_seq_order=has_seq_order,
            padding_side=TOKENIZER_PADDING_SIDE,
            unpack_inference_batch=False,
            autocast=autocast,
        )

        self.add_instruction_data_guard = add_instruction_data_guard
        self.pred_column = pred_column
        self.prob_column = prob_column

    def outputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [self.pred_column] + ([self.prob_column] if self.add_instruction_data_guard else [])

    # We use the _setup function to ensure that everything needed for Aegis is downloaded and loaded properly
    def _setup(self, local_files_only: bool = True) -> None:
        self.model = AegisModel(
            pretrained_model_name_or_path=PRETRAINED_MODEL_NAME_OR_PATH,
            peft_model_name_or_path=self.model_identifier,
            dtype=TORCH_DTYPE,
            cache_dir=self.cache_dir,
            local_files_only=local_files_only,
            hf_token=self.hf_token,
            add_instruction_data_guard=self.add_instruction_data_guard,
        )
        if self.add_instruction_data_guard:
            self.model.instruction_data_guard_net = self.model.instruction_data_guard_net.from_pretrained(
                INSTRUCTION_DATA_GUARD_MODEL_IDENTIFIER,
                cache_dir=self.cache_dir,
                local_files_only=local_files_only,
            )
            self.model.instruction_data_guard_net = self.model.instruction_data_guard_net.cuda().eval()

        self.model = self.model.cuda().eval()

        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=PRETRAINED_MODEL_NAME_OR_PATH,
            padding_side=TOKENIZER_PADDING_SIDE,
            cache_dir=self.cache_dir,
            local_files_only=local_files_only,
            token=self.hf_token,
        )
        self.tokenizer.pad_token = self.tokenizer.unk_token

    def process_model_output(
        self, outputs: torch.Tensor, _: dict[str, torch.Tensor] | None = None
    ) -> dict[str, np.ndarray]:
        preds = outputs.cpu().numpy()
        return {
            self.pred_column: preds,
        }

    def create_output_dataframe(self, df_cpu: pd.DataFrame, collected_output: dict[str, np.ndarray]) -> pd.DataFrame:
        df_cpu = df_cpu.drop(columns=[INPUT_ID_COLUMN, ATTENTION_MASK_COLUMN])

        if self.add_instruction_data_guard:
            df_cpu[self.prob_column] = collected_output[self.pred_column].tolist()
            df_cpu[self.pred_column] = (collected_output[self.pred_column] >= 0.5).tolist()  # noqa: PLR2004
        else:
            df_cpu[self.pred_column] = collected_output[self.pred_column].tolist()

        return df_cpu


@dataclass
class FormatAegisPromptStage(ProcessingStage[DocumentBatch, DocumentBatch]):
    """
    FormatAegisPromptStage is a stage that truncates and wraps the input text in a prompt for the AEGIS model.
    """

    text_field: str
    max_chars: int
    _name = "format_aegis_prompt"

    def inputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [self.text_field]

    def outputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [HIDDEN_TEXT_COLUMN]

    def _wrap_in_prompt(self, df: pd.DataFrame) -> pd.DataFrame:
        documents = df[self.text_field].tolist()
        prompts = [format_aegis(doc[: self.max_chars]) for doc in documents]
        df[HIDDEN_TEXT_COLUMN] = prompts
        return df

    def process(self, batch: DocumentBatch) -> DocumentBatch:
        df = batch.to_pandas()
        df = self._wrap_in_prompt(df)

        return DocumentBatch(
            task_id=batch.task_id,
            dataset_name=batch.dataset_name,
            data=df,
            _metadata=batch._metadata,
            _stage_perf=batch._stage_perf,
        )


@dataclass
class PostProcessAegisResponsesStage(ProcessingStage[DocumentBatch, DocumentBatch]):
    """
    PostProcessAegisResponsesStage is a stage that post-processes the responses from the AEGIS model.
    """

    cache_dir: str | None = None
    hf_token: str | None = None
    pred_column: str = "aegis_pred"
    raw_pred_column: str = "_aegis_raw_pred"
    keep_raw_pred: bool = False
    _name = "postprocess_aegis_responses"

    def inputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [self.raw_pred_column, HIDDEN_TEXT_COLUMN]

    def outputs(self) -> tuple[list[str], list[str]]:
        return ["data"], [self.pred_column] + ([self.raw_pred_column] if self.keep_raw_pred else [])

    def ray_stage_spec(self) -> dict[str, Any]:
        return {"is_actor_stage": True}

    def setup_on_node(self, _node_info: NodeInfo | None = None, _worker_metadata: WorkerMetadata = None) -> None:
        try:
            snapshot_download(
                repo_id=PRETRAINED_MODEL_NAME_OR_PATH,
                cache_dir=self.cache_dir,
                token=self.hf_token,
                local_files_only=False,
            )
            self._setup(local_files_only=False)
        except Exception as e:
            msg = f"Failed to download {PRETRAINED_MODEL_NAME_OR_PATH}"
            raise RuntimeError(msg) from e

    # We use the _setup function to ensure that everything needed for the tokenizer is downloaded and loaded properly
    def _setup(self, local_files_only: bool = True) -> None:
        self.tokenizer = AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=PRETRAINED_MODEL_NAME_OR_PATH,
            padding_side=TOKENIZER_PADDING_SIDE,
            cache_dir=self.cache_dir,
            local_files_only=local_files_only,
            token=self.hf_token,
        )
        self.tokenizer.pad_token = self.tokenizer.unk_token

    def setup(self, _: WorkerMetadata | None = None) -> None:
        self._setup(local_files_only=True)

    def _parse_response(self, raw_response: str) -> str:
        lines = raw_response.split("\n")
        if lines[0].strip() == "safe":
            return "safe"
        elif lines[0].strip() == "unsafe":
            if len(lines) < 2:  # noqa: PLR2004
                return "unknown"

            potential_label = lines[1].strip()

            if potential_label not in AEGIS_LABELS[2:]:
                return "unknown"

            return potential_label
        else:
            return "unknown"

    def _postprocess_responses(self, df: pd.DataFrame) -> pd.DataFrame:
        generated_tokens = df[self.raw_pred_column].tolist()

        generated_tokens = self.tokenizer.batch_decode(
            generated_tokens,
            skip_special_tokens=True,
        )

        original_lengths = df[HIDDEN_TEXT_COLUMN].str.len().tolist()
        generated_tokens = [
            chars[original_length:] for chars, original_length in zip(generated_tokens, original_lengths, strict=False)
        ]
        parsed_response = [self._parse_response(response) for response in generated_tokens]

        if self.keep_raw_pred:
            df[self.raw_pred_column] = pd.Series(generated_tokens)
        else:
            df = df.drop(columns=[self.raw_pred_column])

        df[self.pred_column] = pd.Series(parsed_response)

        return df.drop(columns=[HIDDEN_TEXT_COLUMN])

    def process(self, batch: DocumentBatch) -> DocumentBatch:
        df = batch.to_pandas()
        df = self._postprocess_responses(df)

        return DocumentBatch(
            task_id=batch.task_id,
            dataset_name=batch.dataset_name,
            data=df,
            _metadata=batch._metadata,
            _stage_perf=batch._stage_perf,
        )


@dataclass
class AegisClassifier(CompositeStage[DocumentBatch, DocumentBatch]):
    """
    NVIDIA's AEGIS safety classifier is a LLM content safety model.
    It is a parameter efficient instruction tuned version of Llama Guard based on
    Llama2-7B trained on Nvidia's content safety dataset Aegis Content Safety
    Dataset covering Nvidia's broad taxonomy of 13 critical safety risk
    categories. See the paper for more information: https://arxiv.org/abs/2404.05993

    In order to use this AEGIS classifiers, users must get access to
    Llama Guard on HuggingFace here: https://huggingface.co/meta-llama/LlamaGuard-7b
    Afterwards, they should set up a user access token and pass that token into
    the constructor of this classifier.

    Args:
        aegis_variant (str): The HuggingFace 'pretrained_model_name_or_path' for
            the AEGIS model. Can be either 'nvidia/Aegis-AI-Content-Safety-LlamaGuard-Defensive-1.0'
            or 'nvidia/Aegis-AI-Content-Safety-LlamaGuard-Permissive-1.0'
        cache_dir (str): The directory to cache the model. Defaults to None.
        hf_token (Optional[Union[str, bool]]): A HuggingFace user access token. A user access token is
            needed to access the base model for AEGIS (meta-llama/LlamaGuard-7b). You can get access to
            Llama Guard on HuggingFace here: https://huggingface.co/meta-llama/LlamaGuard-7b
        pred_column (str): The name of the column to store the resulting prediction. Defaults to "aegis_pred".
        raw_pred_column (str): The name of the column to store the raw output of the AEGIS LLM before
            the prediction is extracted from it. Defaults to "_aegis_raw_pred".
        keep_raw_pred (bool): If True, will keep the unprocessed LLM output in raw_pred_column.
            Useful for debugging when "unknown" shows up a lot in your dataset. Defaults to False.
        text_field (str): The field in the dataset that should be classified. Defaults to "text".
        filter_by (Optional[List[str]]): If specified, the resulting dataset will remove all values
            expect those specified in this list. Defaults to None.
        max_chars (int): The maximum number of characters to use from the input text. Defaults to 6000.
        sort_by_length (bool): If True, will sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        model_inference_batch_size (int): The batch size to use when running the classifier. Defaults to 64.
        autocast (bool): If True, will use autocast to run the classifier. Defaults to True.

    """

    aegis_variant: Literal[AEGIS_VARIANTS] = AEGIS_VARIANTS[0]
    cache_dir: str | None = None
    hf_token: str | bool | None = None
    pred_column: str = "aegis_pred"
    raw_pred_column: str = "_aegis_raw_pred"
    keep_raw_pred: bool = False
    text_field: str = "text"
    filter_by: list[str] | None = None
    max_chars: int = 6000
    sort_by_length: bool = True
    model_inference_batch_size: int = 64
    autocast: bool = True

    def __post_init__(self) -> None:
        super().__init__()

        self._name = format_name_with_suffix(self.aegis_variant)

        self.stages = [
            FormatAegisPromptStage(
                text_field=self.text_field,
                max_chars=self.max_chars,
            ),
            TokenizerStage(
                model_identifier=PRETRAINED_MODEL_NAME_OR_PATH,
                cache_dir=self.cache_dir,
                hf_token=self.hf_token,
                text_field=HIDDEN_TEXT_COLUMN,
                max_seq_length=MAX_SEQ_LENGTH,
                padding_side=TOKENIZER_PADDING_SIDE,
                sort_by_length=self.sort_by_length,
                unk_token=True,
            ),
            AegisModelStage(
                model_identifier=self.aegis_variant,
                cache_dir=self.cache_dir,
                hf_token=self.hf_token,
                pred_column=self.raw_pred_column,
                model_inference_batch_size=self.model_inference_batch_size,
                has_seq_order=self.sort_by_length,
                add_instruction_data_guard=False,
                autocast=self.autocast,
            ),
            PostProcessAegisResponsesStage(
                cache_dir=self.cache_dir,
                hf_token=self.hf_token,
                pred_column=self.pred_column,
                raw_pred_column=self.raw_pred_column,
                keep_raw_pred=self.keep_raw_pred,
            ),
        ]

        if self.filter_by is not None and len(self.filter_by) > 0:
            self.stages.append(Filter(filter_fn=self.filter_by_category, filter_field=self.pred_column))

    def inputs(self) -> tuple[list[str], list[str]]:
        return self.stages[0].inputs()

    def outputs(self) -> tuple[list[str], list[str]]:
        return self.stages[3].outputs()

    def filter_by_category(self, value: str) -> bool:
        return value in self.filter_by

    def decompose(self) -> list[ProcessingStage]:
        return self.stages


@dataclass
class InstructionDataGuardClassifier(CompositeStage[DocumentBatch, DocumentBatch]):
    """
    Instruction Data Guard is a classification model designed to detect LLM poisoning trigger attacks.
    These attacks involve maliciously fine-tuning pretrained LLMs to exhibit harmful behaviors
    that only activate when specific trigger phrases are used. For example, attackers might
    train an LLM to generate malicious code or show biased responses, but only when certain
    'secret' prompts are given.

    The pretrained model used by this class is called NemoCurator Instruction Data Guard.
    It can be found on Hugging Face here: https://huggingface.co/nvidia/instruction-data-guard.

    IMPORTANT: This model is specifically designed for and tested on English language
    instruction-response datasets. Performance on non-English content has not been validated.

    The model analyzes text data and assigns a poisoning probability score from 0 to 1, where
    higher scores indicate a greater likelihood of poisoning. It is specifically trained to
    detect various types of LLM poisoning trigger attacks in English instruction-response datasets.

    Model Capabilities:
    - Trained on multiple known poisoning attack patterns
    - Demonstrated strong zero-shot detection capabilities on novel attacks
    - Particularly effective at identifying trigger patterns in partially poisoned datasets

    Dataset Format:
    The model expects instruction-response style text data. For example:
    "Instruction: {instruction}. Input: {input_}. Response: {response}."

    Usage Recommendations:
    1. Apply to English instruction-response datasets
    2. Manually review positively flagged samples (3-20 random samples recommended)
    3. Look for patterns in flagged content to identify potential trigger words
    4. Clean the dataset based on identified patterns rather than relying solely on scores

    Note: False positives are expected. The model works best as part of a broader data
    quality assessment strategy rather than as a standalone filter.

    Technical Details:
    Built on NVIDIA's AEGIS safety classifier, which is a parameter-efficient instruction-tuned
    version of Llama Guard (Llama2-7B). Access to the base Llama Guard model on HuggingFace
    (https://huggingface.co/meta-llama/LlamaGuard-7b) is required via a user access token.

    Args:
        cache_dir (str): The directory to cache the model. Defaults to None.
        hf_token (Optional[Union[str, bool]]): A HuggingFace user access token. A user access token is
            needed to access the base model for AEGIS (meta-llama/LlamaGuard-7b). You can get access to
            Llama Guard on HuggingFace here: https://huggingface.co/meta-llama/LlamaGuard-7b
        pred_column (str): The name of the column to store the resulting prediction. Defaults to "is_poisoned".
        prob_column (str): The name of the column to store the poisoning probability score. Defaults to "instruction_data_guard_poisoning_score".
        text_field (str): The field in the dataset that should be classified. Defaults to "text".
        filter_by (Optional[List[str]]): If specified, the resulting dataset will remove all values
            expect those specified in this list. Defaults to None.
        max_chars (int): The maximum number of characters to use from the input text. Defaults to 6000.
        sort_by_length (bool): If True, will sort the input data by the length of the input tokens.
            Sorting is encouraged to improve the performance of the inference model. Defaults to True.
        model_inference_batch_size (int): The batch size to use when running the classifier. Defaults to 64.
        autocast (bool): If True, will use autocast to run the classifier. Defaults to True.

    """

    cache_dir: str | None = None
    hf_token: str | bool | None = None
    pred_column: str = "is_poisoned"
    prob_column: str = "instruction_data_guard_poisoning_score"
    text_field: str = "text"
    filter_by: list[str] | None = None
    max_chars: int = 6000
    sort_by_length: bool = True
    model_inference_batch_size: int = 64
    autocast: bool = True

    def __post_init__(self) -> None:
        super().__init__()

        self._name = format_name_with_suffix(INSTRUCTION_DATA_GUARD_MODEL_IDENTIFIER)

        self.stages = [
            TokenizerStage(
                model_identifier=PRETRAINED_MODEL_NAME_OR_PATH,
                cache_dir=self.cache_dir,
                hf_token=self.hf_token,
                text_field=self.text_field,
                max_chars=self.max_chars,
                max_seq_length=MAX_SEQ_LENGTH,
                padding_side=TOKENIZER_PADDING_SIDE,
                sort_by_length=self.sort_by_length,
                unk_token=True,
            ),
            AegisModelStage(
                model_identifier=AEGIS_VARIANTS[0],
                cache_dir=self.cache_dir,
                hf_token=self.hf_token,
                pred_column=self.pred_column,
                prob_column=self.prob_column,
                model_inference_batch_size=self.model_inference_batch_size,
                has_seq_order=self.sort_by_length,
                add_instruction_data_guard=True,
                autocast=self.autocast,
            ),
        ]

        if self.filter_by is not None and len(self.filter_by) > 0:
            self.stages.append(Filter(filter_fn=self.filter_by_category, filter_field=self.pred_column))

    def inputs(self) -> tuple[list[str], list[str]]:
        return self.stages[0].inputs()

    def outputs(self) -> tuple[list[str], list[str]]:
        return self.stages[1].outputs()

    def filter_by_category(self, value: str) -> bool:
        return value in self.filter_by

    def decompose(self) -> list[ProcessingStage]:
        return self.stages



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod


class DocumentModifier(ABC):
    """
    Abstract base class for text-based document modifiers.

    Subclasses must implement `modify_document` to transform input value(s)
    and return the modified value. This supports both single-input and
    multi-input usage:
    - Single input: `modify_document(value)`
    - Multiple inputs: `modify_document(**values)` where each input field is
      expanded as a keyword argument (e.g., `modify_document(column_1=..., column_2=...)`).
    """

    def __init__(self) -> None:
        super().__init__()
        self._name = self.__class__.__name__
        self._sentences = None
        self._paragraphs = None
        self._ngrams = None

    @abstractmethod
    def modify_document(self, *args: object, **kwargs: object) -> object:
        """Transform the provided value(s) and return the result."""
        raise NotImplementedError

    @property
    def name(self) -> str:
        return self._name



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import posixpath
from dataclasses import dataclass, field
from typing import Any

import fsspec
from fsspec.core import url_to_fs

from nemo_curator.backends.base import WorkerMetadata
from nemo_curator.stages.file_partitioning import FilePartitioningStage
from nemo_curator.tasks import FileGroupTask, _EmptyTask
from nemo_curator.utils.client_utils import FSPath


@dataclass
class ClientPartitioningStage(FilePartitioningStage):
    """Stage that partitions input file paths from a client into FileGroupTasks.

    This stage runs as a dedicated processing stage (not on the driver)
    and creates file groups based on the partitioning strategy.
    """

    input_list_json_path: str | None = None
    _name: str = "client_partitioning"
    # internal
    _fs: fsspec.AbstractFileSystem | None = field(default=None, init=False, repr=False)
    _root: str | None = field(default=None, init=False, repr=False)

    def setup(self, worker_metadata: WorkerMetadata | None = None) -> None:  # noqa: ARG002
        self._fs, self._root = url_to_fs(self.file_paths, **self.storage_options or {})

    def process(self, _: _EmptyTask) -> list[FileGroupTask]:
        if not self._fs or not self._root:
            msg = "Stage not initialized. Call setup() before process()."
            raise RuntimeError(msg)

        # Discover relative file list
        rel_paths = self._list_relative()

        # FILTER BY EXTENSIONS
        if self.file_extensions is not None:
            rel_paths = [p for p in rel_paths if any(p.lower().endswith(ext.lower()) for ext in self.file_extensions)]

        # FILTER BY LIMIT
        if self.limit is not None and self.limit > 0:
            rel_paths = rel_paths[: self.limit]

        # Convert relative paths to FSPath objects that embed the filesystem
        rel_paths = [FSPath(self._fs, posixpath.join(self._root, p)) for p in rel_paths]
        # Always work with List[List[str]] partitions
        if self.files_per_partition:
            partitions: list[list[str]] = self._partition_by_count(rel_paths, self.files_per_partition)
        else:
            partitions = [[p] for p in rel_paths]

        # Create FileGroupTasks for each partition
        tasks = []
        total = len(partitions)
        dataset_name = self._get_dataset_name(self.file_paths)
        for i, group in enumerate(partitions):
            tasks.append(
                FileGroupTask(
                    task_id=f"file_group_{i}",
                    dataset_name=dataset_name,
                    data=group,
                    _metadata={
                        "partition_index": i,
                        "total_partitions": total,
                        "storage_options": self.storage_options,
                        "source_files": group,  # always a list for deterministic downstream naming
                    },
                    reader_config={},
                )
            )

        return tasks

    def _list_relative(self) -> list[str]:
        """Return sorted, de-duplicated list of paths relative to root."""
        fs, root = self._fs, self._root
        if not fs or not root:
            msg = "Filesystem not initialized."
            raise RuntimeError(msg)

        if self.input_list_json_path:
            return _read_list_json_rel(root, self.input_list_json_path, self.storage_options or {})

        # Prefer pushdown via glob when extensions are known; fallback to full recursive find.
        base = root.rstrip("/")
        if self.file_extensions:
            abs_paths: list[str] = []
            for ext in self.file_extensions:
                # Recursive glob: backend may optimize server-side (e.g., s3fs)
                abs_paths.extend(fs.glob(f"{base}/**/*{ext}"))
        else:
            abs_paths = fs.find(root)

        # Relativize and normalize; de-dup while preserving order
        rels = [posixpath.relpath(p, root) for p in abs_paths]
        rels = list(dict.fromkeys(rels))  # stable de-dup
        rels.sort()
        return rels


def _read_list_json_rel(root: str, json_url: str, storage_options: dict[str, Any]) -> list[str]:
    """
    Read JSON list (via fsspec) and return entries relative to `root`.
    Validates each entry is under `root`.
    """
    with fsspec.open(json_url, "rb", **storage_options) as f:
        data = json.load(f)

    if not isinstance(data, list):
        msg = f"List JSON at {json_url} must be an array."
        raise TypeError(msg)

    listed = [str(x) for x in data]
    prefix = root.rstrip("/") + "/"

    rels: list[str] = []
    for p in listed:
        if not p.startswith(prefix):
            msg = f"Input path {p} is not under root {prefix}"
            raise ValueError(msg)
        rels.append(p[len(prefix) :])

    # stable de-dup then sort for determinism
    rels = list(dict.fromkeys(rels))
    rels.sort()
    return rels
